from dsw_sdk.common.attributes import (
    AttributesMixin,
    BoolAttribute,
    StringAttribute,
)


OWNER_GROUP_MEMBERSHIP = 'OwnerGroupMembershipType'
MEMBER_GROUP_MEMBERSHIP = 'MemberGroupMembershipType'
GROUP_MEMBERSHIP_TYPES = (OWNER_GROUP_MEMBERSHIP, MEMBER_GROUP_MEMBERSHIP)


class UserSuggestion(AttributesMixin):
    first_name = StringAttribute()
    gravatar_hash = StringAttribute()
    image_url = StringAttribute(nullable=True)
    last_name = StringAttribute()
    uuid = StringAttribute()


class UserCreateDTO(AttributesMixin):
    affiliation: str = StringAttribute(nullable=True)  # type: ignore
    email: str = StringAttribute()  # type: ignore
    first_name: str = StringAttribute()  # type: ignore
    last_name: str = StringAttribute()  # type: ignore
    password: str = StringAttribute()  # type: ignore
    role: str = StringAttribute(nullable=True)  # type: ignore


class UserChangeDTO(AttributesMixin):
    active: bool = BoolAttribute()
    affiliation: str = StringAttribute(nullable=True)
    email: str = StringAttribute()
    first_name: str = StringAttribute()
    last_name: str = StringAttribute()
    role: str = StringAttribute()


class GroupMembership(AttributesMixin):
    group_id: str = StringAttribute(immutable=True)
    type: str = StringAttribute(immutable=True, choices=GROUP_MEMBERSHIP_TYPES)



"""
Module containing the main class of the whole library. This is the class
that users will be interacting with. It's the main entrypoint for all
other actions.
"""
from __future__ import annotations

import sys
import logging
from typing import Optional, Type, Union

import requests

from dsw_sdk.config.config import Config
from dsw_sdk.high_level_api.app_config_api import AppConfigAPI
from dsw_sdk.high_level_api.document_api import DocumentAPI
from dsw_sdk.high_level_api.package_api import PackageAPI
from dsw_sdk.high_level_api.questionnaire_api import QuestionnaireAPI
from dsw_sdk.high_level_api.template_api import TemplateAPI
from dsw_sdk.high_level_api.user_api import UserAPI
from dsw_sdk.http_client.interface import HttpClient
from dsw_sdk.http_client.requests_impl.auth import (
    BodyTokenRetrievalStrategy,
    JWTBearerAuth,
)
from dsw_sdk.http_client.requests_impl.http_client import SessionHttpClient
from dsw_sdk.low_level_api.api import LowLevelAPI


HTTP_CLIENT = Union[HttpClient, Type[HttpClient]]  # pylint: disable=C0103


class DataStewardshipWizardSDK:  # pylint: disable=R0902
    """
    This class provides simple and concise way to communicate with the Data
    Stewardship Wizard API. It offers both low-level and object-oriented
    interfaces.

    Low-level API reflects exactly 1:1 the whole SDK API. For each endpoint
    and each HTTP method there's a function on the low-level interface, which
    is accessible as the :attr:`api` attribute.
    This interface is intended for use cases which are not covered by the
    high-level API, offering 100% of the DSW's functionality.

    High-level object-oriented interface is available for the most common
    operations. It's possible that new features will be added in the future.
    This interface is divided by the entities it is concerned with. Right
    now there are 6 interfaces of this kind, accessible via attributes:

        * :attr:`app_config`
        * :attr:`documents`
        * :attr:`packages`
        * :attr:`questionnaires`
        * :attr:`templates`
        * :attr:`users`

    each containing it's own set of functionality (but in most cases it's
    just CRUD operations).

    There some dependency injection parameters you can pass in order to
    alter the default behavior:

        * ``session`` -- you can pre-configure your own session and pass
          it as the argument; it will be used with the HTTP client
        * ``http_client`` -- either instance or class implementing the
          :class:`~interface.HttpClient` interface. If you pass only
          the class, it will get instantiated with all the other config
          values and auth settings. This way you can override some aspects
          of the default HTTP client, but don't have to initialize it
          yourself. But you can also pass the ready-to-use instance and
          all the HTTP client related config will be ignored.
        * ``logger`` -- if you want, you can specify your own logger.

    All configuration values can be set in 3 different ways:

        * passing as keyword arguments to the :meth:`__init__` method
        * setting values as env variables (prefixed with ``DSW_SDK_``)
        * via config file (in YAML format); path is passed in the
          ``conf_file`` keyword argument

    Here is a list of all possible configuration values:

    .. _config_values:

    :Keyword arguments:
        * **api_url** (`str`):
          URL address of the DSW API you want to connect to. It must contain
          valid url scheme (e.g. `https://`) Mandatory if not passing your
          own ``http_client``.
        * **email** (`str`):
          e-mail address of the user on whose behalf you will be acting.
          Mandatory if not passing your own ``http_client``.
        * **password** (`str`):
          password for logging in. Mandatory if not passing your own
          ``http_client``.
        * **enable_ssl** (`bool`):
          flag defining if the HTTP communication should be encrypted or not.
          Default: ``True``.
        * **auth_endpoint** (`str`):
          endpoint on the DSW API that is responsible for the authorization.
          Default: ``'/tokens'``.
        * **headers** (`Dict[str, str]`):
          Dict of default headers to be sent with every HTTP request.
          Default: ``{}``.
        * **default_timeout**:
          This timeout is used as default value when no timeout is specified
          within a specific request. You can pass one numeric value (applied
          for both the connect and read timeouts), tuple of numeric values to
          specify different connect and read timeouts or ``None`` to wait
          forever. Default: ``(6.05, 27)``:
        * **logger_name** (`str`):
          Name of the default logger. Default: ``'dsw_sdk'``.
        * **logger_level** (`Union[int, str]`):
          Logging level of the default logger. You can use both string levels
          (e.g. ``'INFO'``) or integer levels, ideally constants from the
          :mod:`logging` library (e.g. ``logging.INFO``). Default:
          ``logging.WARNING``.
        * **logger_format** (`str`):
          String describing the format of the default logger. For more info,
          consult the official Python docs about :mod:`logging` module.
          Default: ``[%(asctime)s] - %(name)s | %(levelname)s | %(message)s``.

    Example usage:

    .. code-block:: python

        # Basic configuration
        sdk = DataStewardshipWizardSDK(
            api_url='http://localhost:3000',
            email='albert.einstein@example.com',
            password='password',
        )

        # High-level API usage
        temps = sdk.templates.get_templates(size=2, q='my templates')
        for temp in temps:
            # Each `temp` is a `Template` instance
            print(template.usable_packages[0].version)
            temp.name = 'Modified template name'
            temp.save()
    """

    def __init__(
        self,
        session: requests.Session = None,
        http_client: HTTP_CLIENT = None,
        logger: logging.Logger = None,
        **kwargs,
    ):
        """
        :param session: pre-configured session to be used with the
                        HTTP client instead of the default one
        :param http_client: instance or class implementing the
                            :class:`~interface.HttpClient` interface
        :param logger: pre-configured logger to be used
                       instead of the default one
        """
        self._config = Config(**kwargs)
        self.logger = self._init_logger(logger)
        http_client = self._init_http_client(http_client, session)

        self.api = LowLevelAPI(http_client)
        # High-level object-oriented APIs
        self.app_config = AppConfigAPI(self)
        self.documents = DocumentAPI(self)
        self.packages = PackageAPI(self)
        self.questionnaires = QuestionnaireAPI(self)
        self.templates = TemplateAPI(self)
        self.users = UserAPI(self)

    def _init_logger(self, logger: Optional[logging.Logger]) -> logging.Logger:
        if not logger:
            conf = self._config.logger
            conf.validate()

            logger = logging.Logger(conf.logger_name, conf.logger_level)
            formatter = logging.Formatter(conf.logger_format)
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        return logger

    def _init_http_client(
        self,
        http_client: Optional[HTTP_CLIENT],
        session: Optional[requests.Session],
    ) -> HttpClient:
        if http_client and not isinstance(http_client, type):
            # If it is an instance of some object, just return it as it has
            # to be already instantiated and configured from the outside
            return http_client

        self._config.http_client.validate()

        auth_url = (f'{self._config.http_client.api_url}'
                    f'{self._config.http_client.auth_endpoint}')
        credentials = {'email': self._config.http_client.email,
                       'password': self._config.http_client.password}

        return (http_client or SessionHttpClient)(  # type: ignore[operator]
            self._config.http_client.api_url,
            JWTBearerAuth(BodyTokenRetrievalStrategy(auth_url, credentials)),
            self.logger,
            session=session,
            **self._config.http_client,
        )



from dsw_sdk.common.attributes import (
    Attribute,
    AttributesMixin,
    DateTimeAttribute,
    IntegerAttribute,
    ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import MappingType, ObjectType, StringType
from dsw_sdk.high_level_api.dto.user import UserSuggestion


PRIVATE_QUESTIONNAIRE = 'PrivateQuestionnaire'
VISIBLE_EDIT_QUESTIONNAIRE = 'VisibleEditQuestionnaire'
VISIBLE_VIEW_QUESTIONNAIRE = 'VisibleViewQuestionnaire'
QUESTIONNAIRE_VISIBILITIES = (
    PRIVATE_QUESTIONNAIRE,
    VISIBLE_EDIT_QUESTIONNAIRE,
    VISIBLE_VIEW_QUESTIONNAIRE,
)

ANYONE_WITH_LINK_EDIT_QUESTIONNAIRE = 'AnyoneWithLinkEditQuestionnaire'
ANYONE_WITH_LINK_VIEW_QUESTIONNAIRE = 'AnyoneWithLinkViewQuestionnaire'
RESTRICTED_QUESTIONNAIRE = 'RestrictedQuestionnaire'
QUESTIONNAIRE_SHARING = (
    ANYONE_WITH_LINK_EDIT_QUESTIONNAIRE,
    ANYONE_WITH_LINK_VIEW_QUESTIONNAIRE,
    RESTRICTED_QUESTIONNAIRE,
)

DEFAULT_STATE = 'Default'
MIGRATING_STATE = 'Migrating'
OUTDATED_STATE = 'Outdated'
QUESTIONNAIRE_STATES = (
    DEFAULT_STATE,
    MIGRATING_STATE,
    OUTDATED_STATE,
)

ANSWER_REPLY = 'AnswerReply'
INTEGRATION_REPLY = 'IntegrationReply'
ITEM_LIST_REPLY = 'ItemListReply'
MULTI_CHOICE_REPLY = 'MultiChoiceReply'
STRING_REPLY = 'StringReply'
QUESTIONNAIRE_REPLIES = (
    ANSWER_REPLY,
    INTEGRATION_REPLY,
    ITEM_LIST_REPLY,
    MULTI_CHOICE_REPLY,
    STRING_REPLY,
)

GROUP_MEMBER = 'GroupMember'
USER_MEMBER = 'UserMember'
MEMBER_TYPES = (GROUP_MEMBER, USER_MEMBER)

CLEAR_REPLY_EVENT = 'ClearReplyEvent'
SET_LABELS_EVENT = 'SetLabelsEvent'
SET_PHASE_EVENT = 'SetPhaseEvent'
SET_REPLY_EVENT = 'SetReplyEvent'
EVENT_TYPES = (
    CLEAR_REPLY_EVENT,
    SET_LABELS_EVENT,
    SET_PHASE_EVENT,
    SET_REPLY_EVENT,
)

ANSWERED_INDICATION = 'AnsweredIndication'
PHASES_ANSWERED_INDICATION = 'PhasesAnsweredIndication'
INDICATIONS_TYPES = (ANSWERED_INDICATION, PHASES_ANSWERED_INDICATION)

INTEGRATION_TYPE = 'IntegrationType'
PLAIN_TYPE = 'PlainType'
INTEGRATION_REPLY_TYPES = (INTEGRATION_TYPE, PLAIN_TYPE)


class IntegrationReplyType(AttributesMixin):
    value = StringAttribute()
    type = StringAttribute(choices=INTEGRATION_REPLY_TYPES)


class PlainType(IntegrationReplyType):
    pass


class IntegrationType(IntegrationReplyType):
    id = StringAttribute()


class ReplyValue(AttributesMixin):
    type = StringAttribute(choices=QUESTIONNAIRE_REPLIES)


class StringReply(ReplyValue):
    value = StringAttribute()


class AnswerReply(ReplyValue):
    value = StringAttribute()


class MultiChoiceReply(ReplyValue):
    value = ListAttribute(StringType())


class ItemListReply(ReplyValue):
    value = ListAttribute(StringType())


class IntegrationReply(ReplyValue):
    value = Attribute(MappingType('type', {
        PLAIN_TYPE: ObjectType(PlainType),
        INTEGRATION_TYPE: ObjectType(IntegrationType),
    }))


class Reply(AttributesMixin):
    created_at = DateTimeAttribute()
    created_by = ObjectAttribute(UserSuggestion)
    value = Attribute(MappingType('type', {
        ANSWER_REPLY: ObjectType(AnswerReply),
        INTEGRATION_REPLY: ObjectType(IntegrationReply),
        ITEM_LIST_REPLY: ObjectType(ItemListReply),
        MULTI_CHOICE_REPLY: ObjectType(MultiChoiceReply),
        STRING_REPLY: ObjectType(StringReply),
    }))


class GroupMember(AttributesMixin):
    gid = StringAttribute()
    name = StringAttribute()
    type = StringAttribute(choices=MEMBER_TYPES)


class UserMember(UserSuggestion):
    type = StringAttribute(choices=MEMBER_TYPES)


class QuestionnairePermRecordDTO(AttributesMixin):
    member = Attribute(MappingType('type', {
        GROUP_MEMBER: ObjectType(GroupMember),
        USER_MEMBER: ObjectType(UserMember),
    }))
    perms = ListAttribute(StringType())


class QuestionnaireVersion(AttributesMixin):
    created_at = DateTimeAttribute()
    created_by = ObjectAttribute(UserSuggestion)
    description = StringAttribute(nullable=True)
    event_uuid = StringAttribute()
    name = StringAttribute()
    updated_at = DateTimeAttribute()
    uuid = StringAttribute()


class QuestionnaireEvent(AttributesMixin):
    created_at = DateTimeAttribute()
    created_by = ObjectAttribute(UserSuggestion, nullable=True)
    type = StringAttribute(choices=EVENT_TYPES)
    uuid = StringAttribute()


class SetReplyEvent(QuestionnaireEvent):
    path = StringAttribute(nullable=True)
    value = Attribute(MappingType('type', {
        ANSWER_REPLY: ObjectType(AnswerReply),
        INTEGRATION_REPLY: ObjectType(IntegrationReply),
        ITEM_LIST_REPLY: ObjectType(ItemListReply),
        MULTI_CHOICE_REPLY: ObjectType(MultiChoiceReply),
        STRING_REPLY: ObjectType(StringReply),
    }), nullable=True)


class ClearReplyEvent(QuestionnaireEvent):
    path = StringAttribute()


class SetPhaseEvent(QuestionnaireEvent):
    phase = StringAttribute(nullable=True)


class SetLabelsEvent(QuestionnaireEvent):
    path = StringAttribute()
    value = ListAttribute(StringType())


class Indication(AttributesMixin):
    answered_questions = IntegerAttribute()
    unanswered_questions = IntegerAttribute()
    indication_type = StringAttribute(choices=INDICATIONS_TYPES)


class AnsweredIndication(Indication):
    pass


class PhasesAnsweredIndication(Indication):
    pass


class QuestionnaireReportDTO(AttributesMixin):
    indications = ListAttribute(MappingType('indication_type', {
        ANSWERED_INDICATION: ObjectType(AnsweredIndication),
        PHASES_ANSWERED_INDICATION: ObjectType(PhasesAnsweredIndication),
    }))


class QuestionnaireSimple(AttributesMixin):
    name = StringAttribute()
    uuid = StringAttribute()


class QuestionnaireCreateDTO(AttributesMixin):
    name = StringAttribute()
    package_id = StringAttribute()
    sharing = StringAttribute(choices=QUESTIONNAIRE_SHARING)
    visibility = StringAttribute(choices=QUESTIONNAIRE_VISIBILITIES)
    tag_uuids = ListAttribute(StringType(), default=[])
    template_id = StringAttribute(nullable=True, default=None)
    format_uuid = StringAttribute(nullable=True, default=None)


class QuestionnaireCreateFromTemplateDTO(AttributesMixin):
    name = StringAttribute()
    questionnaire_uuid = StringAttribute()



from dsw_sdk.common.attributes import (
    AttributesMixin,
    DictAttribute,
    FloatAttribute,
    ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import MappingType, ObjectType, StringType


DATE_QUESTION_VALUE_TYPE = 'DateQuestionValueType'
NUMBER_QUESTION_VALUE_TYPE = 'NumberQuestionValueType'
STRING_QUESTION_VALUE_TYPE = 'StringQuestionValueType'
TEXT_QUESTION_VALUE_TYPE = 'TextQuestionValueType'
QUESTION_VALUE_TYPES = (
    DATE_QUESTION_VALUE_TYPE,
    NUMBER_QUESTION_VALUE_TYPE,
    STRING_QUESTION_VALUE_TYPE,
    TEXT_QUESTION_VALUE_TYPE,
)

INTEGRATION_QUESTION = 'IntegrationQuestion'
LIST_QUESTION = 'ListQuestion'
MULTI_CHOICE_QUESTION = 'MultiChoiceQuestion'
OPTIONS_QUESTION = 'OptionsQuestion'
VALUE_QUESTION = 'ValueQuestion'
QUESTION_TYPES = (
    INTEGRATION_QUESTION,
    LIST_QUESTION,
    MULTI_CHOICE_QUESTION,
    OPTIONS_QUESTION,
    VALUE_QUESTION,
)

RESOURCE_PAGE_REFERENCE = 'ResourcePageReference'
URL_REFERENCE = 'URLReference'
CROSS_REFERENCE = 'CrossReference'
REFERENCE_TYPES = (
    RESOURCE_PAGE_REFERENCE,
    URL_REFERENCE,
    CROSS_REFERENCE,
)


class Chapter(AttributesMixin):
    question_uuids = ListAttribute(StringType())
    text = StringAttribute(nullable=True)
    title = StringAttribute()
    uuid = StringAttribute()


class Question(AttributesMixin):
    expert_uuids = ListAttribute(StringType())
    question_type = StringAttribute(choices=QUESTION_TYPES)
    reference_uuids = ListAttribute(StringType())
    required_phase_uuid = StringAttribute(nullable=True)
    tag_uuids = ListAttribute(StringType())
    text = StringAttribute(nullable=True)
    title = StringAttribute()
    uuid = StringAttribute()


class OptionsQuestion(Question):
    answer_uuids = ListAttribute(StringType())


class MultiChoiceQuestion(Question):
    choice_uuids = ListAttribute(StringType())


class ListQuestion(Question):
    item_template_question_uuids = ListAttribute(StringType())


class ValueQuestion(Question):
    value_type = StringAttribute(choices=QUESTION_VALUE_TYPES)


class IntegrationQuestion(Question):
    props = DictAttribute(StringType(), StringType())


class MetricMeasure(AttributesMixin):
    metric_uuid = StringAttribute()
    measure = FloatAttribute()
    weight = FloatAttribute()


class Answer(AttributesMixin):
    advice = StringAttribute(nullable=True)
    follow_up_uuids = ListAttribute(StringType())
    label = StringAttribute()
    metric_measures = ListAttribute(ObjectType(MetricMeasure))
    uuid = StringAttribute()


class Choice(AttributesMixin):
    label = StringAttribute()
    uuid = StringAttribute()


class Expert(AttributesMixin):
    email = StringAttribute()
    name = StringAttribute()
    uuid = StringAttribute()


class Reference(AttributesMixin):
    uuid = StringAttribute()
    reference_type = StringAttribute(choices=REFERENCE_TYPES)


class ResourcePageReference(Reference):
    short_uuid = StringAttribute()


class URLReference(Reference):
    label = StringAttribute()
    url = StringAttribute()


class CrossReference(Reference):
    description = StringAttribute()
    target_uuid = StringAttribute()


class Integration(AttributesMixin):
    id = StringAttribute()
    item_url = StringAttribute()
    logo = StringAttribute()
    name = StringAttribute()
    props = ListAttribute(StringType())
    request_body = StringAttribute()
    request_headers = DictAttribute(StringType(), StringType())
    request_method = StringAttribute()
    request_url = StringAttribute()
    response_id_field = StringAttribute()
    response_list_field = StringAttribute()
    response_name_field = StringAttribute()
    uuid = StringAttribute()


class Tag(AttributesMixin):
    color = StringAttribute()
    description = StringAttribute(nullable=True)
    name = StringAttribute()
    uuid = StringAttribute()


class KnowledgeModelEntities(AttributesMixin):
    answers = DictAttribute(StringType(), ObjectType(Answer))
    chapters = DictAttribute(StringType(), ObjectType(Chapter))
    choices = DictAttribute(StringType(), ObjectType(Choice))
    experts = DictAttribute(StringType(), ObjectType(Expert))
    integrations = DictAttribute(StringType(), ObjectType(Integration))
    questions = DictAttribute(StringType(), MappingType('question_type', {
        OPTIONS_QUESTION: ObjectType(OptionsQuestion),
        MULTI_CHOICE_QUESTION: ObjectType(MultiChoiceQuestion),
        LIST_QUESTION: ObjectType(ListQuestion),
        VALUE_QUESTION: ObjectType(ValueQuestion),
        INTEGRATION_QUESTION: ObjectType(IntegrationQuestion),
    }))
    references = DictAttribute(StringType(), MappingType('reference_type', {
        RESOURCE_PAGE_REFERENCE: ObjectType(ResourcePageReference),
        URL_REFERENCE: ObjectType(URLReference),
        CROSS_REFERENCE: ObjectType(CrossReference),
    }))
    tags = DictAttribute(StringType(), ObjectType(Tag))


class KnowledgeModel(AttributesMixin):
    chapter_uuids = ListAttribute(StringType())
    entities = ObjectAttribute(KnowledgeModelEntities)
    integration_uuids = ListAttribute(StringType())
    tag_uuids = ListAttribute(StringType())
    uuid = StringAttribute()



"""
Utility methods and helpers.
"""

import re


def to_camel_case(string: str) -> str:
    """
    Converts a string from snake_case to camelCase.

    :param string: string in snake_case notation

    :return: `string` converted to camelCase
    """
    if string.find('_') == -1:
        return string
    return re.sub(
        r'(.)(_)(.)',
        lambda m: f'{m.group(1)}{m.group(3).upper()}',
        string.lower().strip('_')
    )


def to_snake_case(string: str) -> str:
    """
    Converts a string from camelCase to snake_case.

    :param string: string in camelCase notation

    :return: `string` converted to snake_case
    """
    string = re.sub(r'(.)([A-Z][a-z])', r'\1_\2', string)
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', string).lower()


def truncate_long_string(string: str, max_len: int) -> str:
    """
    Truncates the given text from the first newline character. If no
    newline is present in the ``string``, truncate text to ``max_len``
    characters.
    If the text is shorter than ``max_len``, does nothing.

    :param string: text to truncate
    :param max_len: maximal length of the truncated string

    :return: truncated string, enquoted in double quotes, with the
             ``(truncated)`` string appended at the end
    """
    if len(string) <= max_len:
        return string
    index = string.find('\n')
    index = index if (index != -1 and index <= max_len) else max_len
    return f'"{string[:index]}..." (truncated)'



from dsw_sdk.common.attributes import (
    Alias,
    BoolAttribute,
    DateTimeAttribute,
    DictAttribute,
    ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import (
    MappingType,
    ObjectType,
    StringType,
)
from dsw_sdk.high_level_api.dto.common import (
    PackageSimpleDTO,
)
from dsw_sdk.high_level_api.dto.knowledge_model import KnowledgeModel
from dsw_sdk.high_level_api.dto.questionnaire import (
    CLEAR_REPLY_EVENT,
    ClearReplyEvent,
    QUESTIONNAIRE_SHARING,
    QUESTIONNAIRE_STATES,
    QUESTIONNAIRE_VISIBILITIES,
    QuestionnaireCreateDTO,
    QuestionnaireCreateFromTemplateDTO,
    QuestionnairePermRecordDTO,
    QuestionnaireVersion,
    Reply,
    SET_LABELS_EVENT,
    SET_PHASE_EVENT,
    SET_REPLY_EVENT,
    SetLabelsEvent,
    SetPhaseEvent,
    SetReplyEvent,
)
from dsw_sdk.high_level_api.dto.template import TemplateSimple
from dsw_sdk.high_level_api.models.document import Document
from dsw_sdk.high_level_api.models.model import ListOfModelsAttribute, Model
from dsw_sdk.high_level_api.models.templates.template import (
    TemplateFormat,
)


class Questionnaire(Model):
    created_at = DateTimeAttribute()
    creator_uuid = StringAttribute(nullable=True)
    description = StringAttribute(nullable=True)
    events = ListAttribute(MappingType('type', {
        SET_REPLY_EVENT: ObjectType(SetReplyEvent),
        CLEAR_REPLY_EVENT: ObjectType(ClearReplyEvent),
        SET_PHASE_EVENT: ObjectType(SetPhaseEvent),
        SET_LABELS_EVENT: ObjectType(SetLabelsEvent),
    }))
    format = ObjectAttribute(TemplateFormat, nullable=True)
    format_uuid = StringAttribute(nullable=True)
    is_template = BoolAttribute()
    knowledge_model = ObjectAttribute(KnowledgeModel)
    labels = DictAttribute(StringType(), StringType())
    phase_uuid = StringAttribute()
    name = StringAttribute()
    package = ObjectAttribute(PackageSimpleDTO)
    package_id = StringAttribute()
    permissions = ListAttribute(ObjectType(QuestionnairePermRecordDTO))
    replies = DictAttribute(StringType(), ObjectType(Reply))
    selected_tag_uuids = ListAttribute(StringType())
    sharing = StringAttribute(choices=QUESTIONNAIRE_SHARING)
    state = StringAttribute(choices=QUESTIONNAIRE_STATES)
    # TODO: Unite with `selected_tag_uuids`
    tag_uuids = Alias('selected_tag_uuids')
    template = ObjectAttribute(TemplateSimple, nullable=True)
    template_id = StringAttribute(nullable=True)
    updated_at = DateTimeAttribute()
    versions = ListAttribute(ObjectType(QuestionnaireVersion))
    visibility = StringAttribute(choices=QUESTIONNAIRE_VISIBILITIES)

    documents = ListOfModelsAttribute(Document, default=[])

    def _create(self):
        dto = QuestionnaireCreateDTO(**self.attrs())
        dto.validate()
        data = self._sdk.api.post_questionnaires(body=dto.to_json()).json()
        detail_data = self._sdk.api.get_questionnaire(
            qtn_uuid=data['uuid']
        ).json()
        self._update_attrs(**detail_data)

    def create_from_template(self, **kwargs):
        dto = QuestionnaireCreateFromTemplateDTO(**kwargs)
        dto.validate()
        created_data = self._sdk.api.post_questionnaires_from_template(
            body=dto.to_json()
        ).json()
        data = self._sdk.api.get_questionnaire(
            qtn_uuid=created_data['uuid']
        ).json()
        self._update_attrs(**data)

    def _update(self):
        raise NotImplementedError('Cannot update questionnaires')

    def _delete(self):
        raise NotImplementedError('Cannot delete questionnaires')



from typing import Any

from dsw_sdk.common.attributes import AttributesMixin, StringAttribute
from dsw_sdk.common.utils import truncate_long_string
from dsw_sdk.high_level_api.models.model import Model


class TemplateFileChangeDTO(AttributesMixin):
    content = StringAttribute()
    file_name = StringAttribute()

    template_id = StringAttribute(immutable=True)


class TemplateFile(Model):
    content = StringAttribute()
    file_name = StringAttribute()

    template_id = StringAttribute(immutable=True)

    def _attr_to_str(self, name: str, value: Any) -> str:
        if name == 'content':
            return truncate_long_string(value, 50)
        return super()._attr_to_str(name, value)

    def _create(self):
        dto = TemplateFileChangeDTO(**self.attrs())
        data = self._sdk.api.post_template_files(
            dto.template_id,
            body=dto.to_json(),
        ).json()
        self._update_attrs(**data)

    def _update(self):
        dto = TemplateFileChangeDTO(**self.attrs())
        data = self._sdk.api.put_template_file(
            dto.template_id,
            self.uuid,
            body=dto.to_json(),
        ).json()
        self._update_attrs(**data)

    def _delete(self):
        self._sdk.api.delete_template_file(self.template_id, self.uuid)



from dsw_sdk.common.attributes import (
    AttributesMixin,
    DateTimeAttribute,
    DictAttribute,
    IntegerAttribute,
    ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import ObjectType, StringType
from dsw_sdk.high_level_api.dto.common import (
    OrganizationSimple,
    PackageSimpleDTO,
)


OUTDATED_TEMPLATE_STATE = 'OutdatedTemplateState'
UNKNOWN_TEMPLATE_STATE = 'UnknownTemplateState'
UNPUBLISHED_TEMPLATE_STATE = 'UnpublishedTemplateState'
UP_TO_DATE_TEMPLATE_STATE = 'UpToDateTemplateState'
UNSUPPORTED_METAMODEL_VERSION_TEMPLATE_STATE = ('UnsupportedMetamodel'
                                                'VersionTemplateState')
TEMPLATE_STATES = (
    OUTDATED_TEMPLATE_STATE,
    UNKNOWN_TEMPLATE_STATE,
    UNPUBLISHED_TEMPLATE_STATE,
    UNSUPPORTED_METAMODEL_VERSION_TEMPLATE_STATE,
    UP_TO_DATE_TEMPLATE_STATE,
)


class TemplateAllowedPackage(AttributesMixin):
    km_id = StringAttribute(nullable=True)
    max_version = StringAttribute(nullable=True)
    min_version = StringAttribute(nullable=True)
    org_id = StringAttribute(nullable=True)


class TemplateFormatStep(AttributesMixin):
    name = StringAttribute()
    options = DictAttribute(StringType(), StringType())


class TemplateFormat(AttributesMixin):
    color = StringAttribute()
    icon = StringAttribute()
    name = StringAttribute()
    short_name = StringAttribute()
    steps = ListAttribute(ObjectType(TemplateFormatStep), nullable=True)
    uuid = StringAttribute()


class TemplateChangeDTO(AttributesMixin):
    allowed_packages = ListAttribute(ObjectType(TemplateAllowedPackage))
    description = StringAttribute()
    formats = ListAttribute(ObjectType(TemplateFormat))
    license = StringAttribute()
    metamodel_version = IntegerAttribute()
    name = StringAttribute()
    organization_id = StringAttribute()
    readme = StringAttribute()
    recommended_package_id = StringAttribute(nullable=True)
    template_id = StringAttribute()
    version = StringAttribute()


class TemplateSimple(AttributesMixin):
    description = StringAttribute()
    formats = ListAttribute(ObjectType(TemplateFormat))
    id = StringAttribute()
    name = StringAttribute()
    version = StringAttribute()


class TemplateSimpleDTO(TemplateSimple):
    allowed_packages = ListAttribute(ObjectType(TemplateAllowedPackage))
    created_at = DateTimeAttribute(read_only=True)
    license = StringAttribute()
    metamodel_version = IntegerAttribute()
    organization = ObjectAttribute(OrganizationSimple, nullable=True,
                                   read_only=True)
    organization_id = StringAttribute()
    readme = StringAttribute()
    recommended_package_id = StringAttribute(nullable=True)
    state = StringAttribute(choices=TEMPLATE_STATES, read_only=True)
    template_id = StringAttribute()
    usable_packages = ListAttribute(ObjectType(PackageSimpleDTO),
                                    read_only=True)



"""
App config high-level API.
"""

from __future__ import annotations

from dsw_sdk.high_level_api.common_api import API
from dsw_sdk.high_level_api.models.app_config import AppConfig


class AppConfigAPI(API):
    """
    API for the App config "entity".

    There is just one GET and one PUT method on the DSW API, so we can
    only return the app config values, updated it and then save it on
    the remote server once again.
    """
    model_class = AppConfig

    def get_config(self) -> AppConfig:
        """
        Get app config from the server.

        :return: object containing DSW app config
        """
        config_data = self._sdk.api.get_configs_app().json()
        return AppConfig(self._sdk, __update_attrs=config_data)



"""
Packages high-level API.
"""

from typing import List

from dsw_sdk.high_level_api.common_api import API, RegistryAPIMixin


class PackageAPI(API, RegistryAPIMixin):
    """
    API for the Package entities.

    This class provides a method for pulling new versions of packages
    (knowledge models). Other methods might be added in the future.

    Example usage:

    .. code-block:: python

        api = PackageAPI(...)

        # Pull specified packages from the Data Stewardship Registry
        api.pull_packages(['dsw:package1:2.2.0', 'dsw:package2:1.0.0'])
    """

    def pull_packages(self, ids: List[str]):
        """
        Pulls given packages from the Data Stewardship Registry, so they
        become available in the DSW instance.

        :param ids: IDs of the packages you want to pull form the registry
        """
        self._pull(self._sdk.api.post_package_pull, 'Packages', ids)

    def update_packages(self, ids: List[str] = None):
        # This will be implemented in future; we need to define the package
        # model first.
        raise NotImplementedError('Updating packages is not yet implemented')



"""
Documents high-level API.
"""
from __future__ import annotations

from typing import List

from dsw_sdk.high_level_api.common_api import API
from dsw_sdk.high_level_api.models.document import Document
from dsw_sdk.http_client.interface import NotFoundError


class DocumentAPI(API):
    """
    API for the Document entities.

    For now, there are just two methods for getting one or many documents
    (as others were not yet required, but might be implemented in the future).

    Example usage:

    .. code-block:: python

        api = DocumentAPI(...)

        # Get one document by UUID
        doc = api.get_document('some-uuid-1234')

        # Get all documents for a given questionnaire
        docs = api.get_documents(questionnaire_uuid='q-uuid-5')

        # Get page number 1 (each page having 10 documents) of documents
        # containing the "foo" string, sorted by the UUID attribute in the
        # ascending order
        docs = api.get_documents(q='foo', page=1, size=10, sort='uuid,asc')
    """
    model_class = Document

    def get_document(self, uuid: str) -> Document:
        """
        Retrieves one document, identified by it's UUID.

        :param uuid: universally unique identifier of the document

        :return: object representing a document
        """
        res = self.get_documents()
        for document in res:
            if document.uuid == uuid:
                return document
        raise NotFoundError(self._last_response)

    def get_documents(self, questionnaire_uuid: str = None,
                      **query_params) -> List[Document]:
        """
        Retrieves list of documents.
        Possibly identified by the questionnaire to which they belong to.

        :param questionnaire_uuid: UUID of the questionnaire whose documents
            we want to fetch
        :param query_params: optional query params: ``q``, ``size``, ``page``
            and ``sort``

        :return: list of objects, each representing a document
        """
        if questionnaire_uuid:
            query_params['questionnaire_uuid'] = questionnaire_uuid
        documents = self._get_many(self._sdk.api.get_documents,
                                   'documents', **query_params)
        return documents



from typing import Any

from dsw_sdk.common.attributes import AttributesMixin, StringAttribute
from dsw_sdk.common.utils import truncate_long_string
from dsw_sdk.high_level_api.models.model import Model


class TemplateAssetCreateDTO(AttributesMixin):
    content = StringAttribute()
    content_type = StringAttribute(nullable=True)
    file_name = StringAttribute()

    template_id = StringAttribute(immutable=True)

    def files(self):
        file_tuple = (self.file_name, self.content)
        if self.content_type:
            file_tuple += (self.content_type,)
        return {'file': file_tuple}


class TemplateAsset(Model):
    content = StringAttribute()
    content_type = StringAttribute()
    file_name = StringAttribute()

    template_id = StringAttribute(immutable=True)

    def _attr_to_str(self, name: str, value: Any) -> str:
        if name == 'content':
            return truncate_long_string(value, 50)
        return super()._attr_to_str(name, value)

    def _create(self):
        dto = TemplateAssetCreateDTO(**self.attrs())
        data = self._sdk.api.post_template_assets(
            dto.template_id,
            files=dto.files(),
        ).json()
        self._update_attrs(**data)

    def _update(self):
        raise NotImplementedError('You cannot edit a template asset.')

    def _delete(self):
        self._sdk.api.delete_template_asset(self.template_id, self.uuid)



"""
Module containing a synchronous implementation of the
:class:`~interface.HttpClient`
interface via the popular `requests` library.
"""

from __future__ import annotations

import logging
import time
from typing import Any, Callable, Dict, Tuple, Union

import requests
from requests.auth import AuthBase

from dsw_sdk.http_client.interface import (
    BadRequestError,
    ForbiddenError,
    HttpClient,
    HttpError,
    HttpResponse,
    NotFoundError,
)


class RequestsHttpResponse(HttpResponse):
    """
    Requests implementation of the :class:`~interface.HttpResponse` interface.
    """

    def __init__(self, response: requests.Response):
        """
        :param response: original :class:`requests.Response` object
        """
        self._response = response

    @property
    def path(self) -> str:
        return self._response.request.path_url

    @property
    def orig_response(self) -> Any:
        return self._response

    @property
    def text(self) -> str:
        return self._response.text

    def json(self, **kwargs) -> Dict[str, Any]:
        return self._response.json(**kwargs)


class SessionHttpClient(HttpClient):
    """
    HTTP client for easy communication with the Data Stewardship Wizard API.
    It's a synchronous implementation of a :class:`~interface.HttpClient`
    interface via the ``Requests`` package.

    Uses :class:`requests.Session` in order to re-use HTTP connections and
    improve overall performance. Also transparently mediates the authentication
    process.

    It is possible to add custom logic with :meth:`before_request` and
    :meth:`after_request` hooks.

    All requests raise :class:`~interface.HttpError`, so you don't have to
    check every response for valid status code. Just assume it's OK and catch
    the aforementioned exception somewhere.
    The :class:`~interface.UnexpectedAuthError` is raised if some unexpected
    error occurs during the authentication process. This means the API
    responded in an unexpected way and something is wrong.

    :Keyword arguments:
        * **default_timeout**:
            This timeout is used as default value when no timeout is specified
            within a request. You can pass one numeric value (applied for both
            the connect and read timeouts), tuple of numeric values to specify
            different connect and read timeouts or ``None`` to wait forever.
            Default: ``None``.
        * **session** (`requests.Session`):
            If you want, you can configure your own session (instance of a
            :class:`requests.Session` class) outside of this client and pass
            it as keyword argument.
        * **headers**: (`Dict[str, str]`)
            Dict of all the HTTP headers you want to send with each requests.
        * **enable_ssl** (`bool`):
            If you set this flag to ``True``, every request gets encrypted and
            the whole communication will be done over SSL/TLS.
            Default: ``True``.
    """

    def __init__(self, base_url: str, auth_class: AuthBase,
                 logger: logging.Logger, **kwargs):
        """
        :param base_url: URL of the DSW API
        :param auth_class: class responsible for the authentication process
        :param logger: logger object
        """
        super().__init__(base_url, auth_class, logger, **kwargs)
        self._base_url = base_url
        self._logger = logger
        self._default_timeout = kwargs.get('default_timeout')

        self._session = kwargs.get('session') or requests.Session()
        self._session.headers.update(kwargs.get('headers') or {})
        if not kwargs.get('session'):
            self._session.verify = kwargs.get('enable_ssl', True)
        if not self._session.auth:
            self._session.auth = auth_class

    def before_request(  # pylint: disable=R0201
        self,
        *args,
        **kwargs
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """
        This method can be overridden by users to provide custom logic before
        sending the request.

        ``args`` and ``kwargs`` are the exact arguments that go into the
        request. This method must return the same, new or modified ``args``
        and ``kwargs``.

        :return: current, new or modified positional and keyword arguments
                 for the request
        """
        return args, kwargs

    def after_request(  # pylint: disable=R0201
        self,
        response: RequestsHttpResponse,
    ) -> RequestsHttpResponse:
        """
        This method can be overridden by users to provide custom logic after
        receiving the response.

        ``response`` is the original response from the server. This
        method must return the same, new or modified response.

        :param response: original response from the server

        :return: current, new or modified response from the server
        """
        return response

    def _request(self, func: Callable, path: Union[str, bytes],
                 *args, **kwargs) -> RequestsHttpResponse:
        """
        General method for performing an HTTP request to the DSW API. Each
        request invokes the ``self._session.auth`` class and thus perform the
        authentication. Also call :meth:`before_request` and
        :meth:`after_request` hooks.

        :param func: function from the :mod:`requests` module to perform the
                     HTTP request
        :param path: path part of the URL

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating

        :return: HTTP response, possibly modified by the
                 :meth:`after_request` hook
        """
        args, kwargs = self.before_request(*args, **kwargs)

        method = func.__name__.upper()
        self._logger.info(f'Starting {method} request to `{path}` with '
                          f'arguments: `{args}` and `{kwargs}`.')
        start_time = time.time()

        try:
            res = func(f'{self._base_url}{path}', *args,
                       timeout=self._default_timeout, **kwargs)
            res.raise_for_status()
        except requests.HTTPError as err:
            resp = RequestsHttpResponse(err.response)
            if err.response.status_code == 400:
                raise BadRequestError(resp)
            if err.response.status_code == 403:
                raise ForbiddenError(resp)
            if err.response.status_code == 404:
                raise NotFoundError(resp)
            raise HttpError(err.response.status_code, resp) from err

        self._logger.info(f'Request finished. Took about '
                          f'{time.time() - start_time:.2f} seconds.')
        res = self.after_request(RequestsHttpResponse(res))
        return res

    def head(self, path: str, **kwargs) -> RequestsHttpResponse:
        """
        Sends a HEAD request. Returns :class:`RequestsHttpResponse` object.

        :param path: path for the new :class:`requests.Request` object.
        :param kwargs: Optional arguments that the request takes.

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating
        """
        return self._request(self._session.head, path, **kwargs)

    def options(self, path: str, **kwargs) -> RequestsHttpResponse:
        """
        Sends an OPTIONS request. Returns :class:`RequestsHttpResponse` object.

        :param path: path for the new :class:`requests.Request` object.
        :param kwargs: Optional arguments that the request takes.

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating
        """
        return self._request(self._session.options, path, **kwargs)

    def get(self, path: str, **kwargs) -> RequestsHttpResponse:
        """
        Sends a GET request. Returns :class:`RequestsHttpResponse` object.

        :param path: path for the new :class:`requests.Request` object.
        :param kwargs: Optional arguments that request takes.

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating
        """
        return self._request(self._session.get, path, **kwargs)

    def post(self, path: str, body: Dict[str, Any] = None,
             **kwargs) -> RequestsHttpResponse:
        """
        Sends a POST request. Returns :class:`RequestsHttpResponse` object.

        :param path: path for the new :class:`requests.Request` object.
        :param body: (optional) json to send in the body of the
                     :class:`requests.Request`.
        :param kwargs: Optional arguments that the request takes.

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating
        """
        return self._request(self._session.post, path, json=body, **kwargs)

    def put(self, path: str, body: Dict[str, Any] = None,
            **kwargs) -> RequestsHttpResponse:
        """
        Sends a PUT request. Returns :class:`RequestsHttpResponse` object.

        :param path: path for the new :class:`requests.Request` object.
        :param body: (optional) json to send in the body of the
                     :class:`requests.Request`.
        :param kwargs: Optional arguments that the request takes.

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating
        """
        return self._request(self._session.put, path, json=body, **kwargs)

    def delete(self, path: str, **kwargs) -> RequestsHttpResponse:
        """
        Sends a DELETE request. Returns :class:`RequestsHttpResponse` object.

        :param path: path for the new :class:`requests.Request` object.
        :param kwargs: Optional arguments that the request takes.

        :raises: :exc:`~interface.HttpError` on HTTP status codes between 400
                 and 599 while authenticating or doing the request itself
        :raises: :exc:`~interface.UnexpectedAuthError` on unexpected errors
                 while authenticating
        """
        return self._request(self._session.delete, path, **kwargs)



"""
Users high-level API.
"""
from __future__ import annotations

from typing import Dict, List

from dsw_sdk.high_level_api.common_api import API
from dsw_sdk.high_level_api.dto.questionnaire import USER_MEMBER
from dsw_sdk.high_level_api.models.questionnaire import Questionnaire
from dsw_sdk.high_level_api.models.user import User


class UserAPI(API):
    """
    API for the User entities.

    Supports all of the CRUD operations.

    Example usage:

    .. code-block:: python

        api = UserAPI(...)

        # Get one user by UUID
        user = api.get_user('some-user-uuid')

        # Get all users
        users = api.get_users()

        # Get page number 1 (each page having 10 users) of users
        # containing the "foo" string, sorted by the UUID attribute in the
        # ascending order
        users = api.get_users(q='foo', page=1, size=10, sort='uuid,asc')

        # Create user
        user = api.create_user(first_name='John', ...)

        # Delete user
        api.delete_user(user.uuid)

        # Delete list of users
        api.delete_users(['some-user-uuid', 'another-user-123'])
    """
    model_class = User

    def _load_questionnaires(
        self,
        user_uuids: List[str],
    ) -> Dict[str, List[Questionnaire]]:
        """
        For a given list of user UUIDs, loads all of their questionnaires and
        associated documents.
        """
        user_questionnaires: Dict[str, List[Questionnaire]] = {
            uuid: [] for uuid in user_uuids
        }

        questionnaires = self._sdk.questionnaires.get_questionnaires()
        for quest in questionnaires:
            for perm in quest.permissions:
                if (
                    perm.member.type == USER_MEMBER
                    and perm.member.uuid in user_uuids
                ):
                    user_questionnaires[perm.member.uuid].append(quest)

        return user_questionnaires

    def get_user(self, uuid: str) -> User:
        """
        Retrieves one user, identified by it's UUID.
        Also loading all his/her related questionnaires and documents.

        :param uuid: user identifier

        :return: object representing a user
        """
        user = self._get_one(self._sdk.api.get_user, uuid)
        questionnaire_lookup = self._load_questionnaires([uuid])
        user._update_attrs(questionnaires=questionnaire_lookup[uuid])
        return user

    def get_users(self, **query_params) -> List[User]:
        """
        Retrieves list of users.
        Also loading all related questionnaires and documents.

        :param query_params: optional query params: ``q``, ``size``, ``page``
            and ``sort``

        :return: list of objects, each representing a user
        """
        users = self._get_many(self._sdk.api.get_users,
                               'users', **query_params)
        user_uuids = [user.uuid for user in users]
        questionnaire_lookup = self._load_questionnaires(user_uuids)
        for user in users:
            user_questionnaires = questionnaire_lookup[user.uuid]
            user._update_attrs(questionnaires=user_questionnaires)
        return users

    def create_user(self, **kwargs) -> User:
        """
        Creates a user with given data on the DSW server.

        :param kwargs: all the data needed for the user creation

        :return: object representing the new user
        """
        return self._create_new(**kwargs)

    def delete_user(self, uuid: str):
        """
        Deletes a user from the DSW server.

        :param uuid: UUID of the user to delete
        """
        self._delete_one(self._sdk.api.delete_user, uuid)

    def delete_users(self, uuids: List[str]):
        """
        Deletes multiple users on the DSW server, identified by UUIDs.

        :param uuids: UUIDs of users to delete
        """
        self._delete_many(self._sdk.api.delete_user, uuids)



"""
Common classes for high-level APIs.
"""

from typing import Any, Callable, Dict, List, Optional, Type

from dsw_sdk.high_level_api.models.model import Model
from dsw_sdk.http_client.interface import (
    BadRequestError,
    HttpResponse,
    NotFoundError,
)


class NotInRegistryError(Exception):
    """
    Raised when attempting to pull an entity that is not present in the
    Data Stewardship Registry.
    """
    msg = '{} {} were not found in the registry.'

    def __init__(self, entity_types: str, ids: List[str]):
        super().__init__(self.msg.format(entity_types, ids))


class API:
    """
    Class defining common logic for all the other high-level APIs.
    Every subclass should define the :attr:`model_class` class attribute (it's
    the model class that the API is working with, e.g. :class:`~models.User`).
    """
    model_class: Type[Model]

    def __init__(self, sdk):
        self._sdk = sdk
        self._last_response: HttpResponse = None

    def _create_loaded(self, data: Dict[str, Any]) -> Model:
        return self.model_class(self._sdk, __update_attrs=data)

    def _get_one(self, func: Callable[[str], HttpResponse],
                 uuid: str) -> Model:
        data = func(uuid).json()
        return self._create_loaded(data)

    def _get_many_data(self, func: Callable[[Dict[str, Any]], HttpResponse],
                       data_key: str, **query_params) -> List[Dict[str, Any]]:
        self._last_response = func(query_params)
        response = self._last_response.json()
        data: List[Dict[str, Any]] = response['_embedded'][data_key]

        if query_params.get('page') is None:
            # If the user haven't specified `page` nor `size`, we can assume
            # that he/she wants all available results. But the API returns
            # at most 100 entries so we have to load to gradually. Same goes
            # for the case, when user specifies a size greater than 100.
            page = 1
            while (
                response['page']['number'] + 1 < response['page']['totalPages']
                and len(data) < query_params.get('size', 2**32)
            ):
                response = func({**query_params, 'page': page}).json()
                page += 1
                data.extend(response['_embedded'][data_key])

        return data

    def _get_many(self, func: Callable[[Dict[str, Any]], HttpResponse],
                  data_key: str, **query_params) -> List[Model]:
        data = self._get_many_data(func, data_key, **query_params)
        return [self._create_loaded(d) for d in data]

    def _create_new(self, **kwargs) -> Model:
        instance = self.model_class(self._sdk, **kwargs)
        instance.save()
        return instance

    def _delete_one(self, func: Callable[[str], HttpResponse], uuid: str):
        func(uuid)

    def _delete_many(self, func: Callable[[str], HttpResponse],
                     uuids: List[str]):
        for uuid in uuids:
            try:
                func(uuid)
            except NotFoundError:
                pass


class RegistryAPIMixin:
    """
    Mixin class providing common functionality for communicating with the
    Data Stewardship Registry.
    """

    @staticmethod
    def _pull(func: Callable[[str], HttpResponse],
              entity_types: str, ids: List[str]):
        not_found_ids = []
        for id_ in ids:
            try:
                func(id_)
            except BadRequestError as err:
                if "wasn't found in Registry" in err.msg:
                    not_found_ids.append(id_)
                elif 'already exists' in err.msg:
                    pass
                else:
                    raise
        if not_found_ids:
            raise NotInRegistryError(entity_types, not_found_ids)

    @staticmethod
    def _update(
        get_one: Callable[[str], Model],
        get_many: Callable[[], List[Model]],
        pull: Callable[[List[str]], None],
        ids: Optional[List[str]],
    ):
        if ids is None:
            entities = get_many()
        else:
            entities = [get_one(id_) for id_ in ids]

        remote_latest_ids = []
        for entity in entities:
            if (
                not entity.remote_latest_version
                or entity.remote_latest_version == entity.version
            ):
                continue
            remote_latest_id = (f'{entity.organization_id}:'
                                f'{entity.template_id}:'
                                f'{entity.remote_latest_version}')
            remote_latest_ids.append(remote_latest_id)

        pull(remote_latest_ids)



"""
Module containing all the data types used with :class:`~common.Attribute`
classes.
"""
from __future__ import annotations

import json
import re
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sized,
    Type as TypingType,
    TypeVar,
    Union,
)

from dsw_sdk.common.utils import to_camel_case


T = TypeVar('T')  # pylint: disable=C0103


class Type:
    """
    The most general type, parent for all specific types.
    """
    _type: TypingType

    def __repr__(self):
        return str(self)

    def __str__(self):
        # This is how the type will be displayed to the users in
        # error messages. Usually we want to display corresponding
        # builtin type or type from the typing module.
        return str(self._type.__name__)

    def _from_string(self, value: str) -> Any:
        """
        Used for converting from string value.

        :param value: string value to be converted to ``self._type`` type

        :return: possibly converted value
        """
        return value

    def validate(self, value: Any):
        """
        Validates if ``value`` is of correct data type.

        :param value: value to be validated

        :raises: :exc:`ValueError` if validation fails
        """
        # Not performing instance check (`isinstance`), but comparing types
        # directly because we want subclasses to fail this check, e.g.
        # `isinstance(True, int)` would pass but `type(True) == int` not
        if not type(value) == self._type:  # pylint: disable=C0123
            raise ValueError

    def convert(self, value: Any) -> Any:
        """
        Tries to convert ``value`` to ``self._type`` data type.

        :param value: value to be converted

        :return: possibly converted value, but it might also
                 be just the original value
        """
        if isinstance(value, str):
            return self._from_string(value)
        return value

    def to_json(self, value: Any) -> Any:
        """
        Converts the ``value`` to JSON.

        Be aware, that the result is not a string, but is instead
        represented with built-in Python types.

        :param value: value to be converted

        :return: JSON representation of the ``value``
        """
        return value

    def value_repr(self, value: Any) -> str:
        """
        Returns the string representation of the ``value`` for
        the ``self._type`` data type.

        :param value: value to be represented as a string

        :return: string representing ``value`` for this particular data type
        """
        return str(value)


class AnyType(Type):
    def __str__(self):
        return 'Any'

    def validate(self, value: Any):
        pass

    def convert(self, value: Any) -> Any:
        return value


class NoneType(Type):
    _type = type(None)

    def _from_string(self, value: str) -> Optional[str]:
        if value.lower() in ('none', 'null'):
            return None
        return value


class BoolType(Type):
    _type = bool

    def _from_string(self, value: str) -> Union[bool, str]:
        if value.lower() in ('true', 'false'):
            return value.lower() == 'true'
        return value


class StringType(Type):
    _type = str

    def convert(self, value: Any) -> str:
        return str(value)

    def value_repr(self, value: str) -> str:
        return f'"{value}"'


class IntegerType(Type):
    _type = int

    def convert(self, value: Any) -> int:
        try:
            # Converting to string first in order
            # not to convert bool or float values
            return int(str(value))
        except ValueError:
            return value


class FloatType(Type):
    _type = float

    def convert(self, value: Any) -> float:
        try:
            # Converting to string first in order
            # not to convert bool values
            return float(str(value))
        except ValueError:
            return value


class UnionType(Type):
    def __init__(self, *of_types: Type):
        self._of_types = of_types

    def __str__(self):
        types = []
        for type_ in self._of_types:
            if isinstance(type_, UnionType):
                types.extend(str(t) for t in type_._of_types)
            else:
                types.append(str(type_))
        return f'Union[{", ".join(types)}]'

    def _iter_over_types(self, value, callback):
        for type_ in self._of_types:
            try:
                return callback(type_, value)
            except ValueError:
                pass
        raise ValueError

    def validate(self, value: Any):
        self._iter_over_types(value, lambda type_, val: type_.validate(val))

    def convert(self, value: Any) -> Any:
        def callback(type_, val):
            converted = type_.convert(val)
            type_.validate(converted)
            return converted

        try:
            return self._iter_over_types(value, callback)
        except ValueError:
            return value

    def to_json(self, value: Any) -> Any:
        def callback(type_, val):
            type_.validate(val)
            return type_.to_json(val)

        return self._iter_over_types(value, callback)

    def value_repr(self, value: Any) -> str:
        def callback(type_, val):
            type_.validate(val)
            return type_.value_repr(val)

        return self._iter_over_types(value, callback)


class TupleType(Type):
    _type = tuple

    def __init__(self, *of_types: Type):
        self._of_types = of_types

    def __str__(self):
        types_str = ', '.join([str(type) for type in self._of_types])
        return f'Tuple[{types_str}]'

    def validate(self, value: Any):
        if not isinstance(value, (list, tuple)):
            raise ValueError
        if len(value) != len(self._of_types):
            raise ValueError
        for val, type_ in zip(value, self._of_types):
            type_.validate(val)

    def _from_string(self, value: str) -> Union[tuple, List[str]]:
        values = value.replace(' ', '').lstrip('(').rstrip(')').split(',')
        if values == ['']:
            return ()
        return values

    def convert(self, value: Any) -> tuple:
        converted = super().convert(value)
        if isinstance(converted, Sized) and isinstance(converted, Iterable):
            if len(self._of_types) != len(converted):
                return value
            res = []
            for val, type_ in zip(converted, self._of_types):
                converted_val = type_.convert(val)
                try:
                    # Checking that each item of the
                    # tuple is converted correctly
                    type_.validate(converted_val)
                except ValueError:
                    # Otherwise we must return the original value
                    return value
                res.append(converted_val)
            return tuple(res)
        return value

    def to_json(self, value: tuple) -> list:
        return [t.to_json(v) for v, t in zip(value, self._of_types)]

    def value_repr(self, value: tuple) -> str:
        if (
            any(isinstance(type_, (ObjectType, MappingType))
                for type_ in self._of_types)
            and value
        ):
            return f'(...) ({len(value)} items)'
        return str(value)


class ListType(Type):
    _type = list

    def __init__(self, of_type: Type = AnyType()):
        self._of_type = of_type

    def __str__(self):
        return f'List[{str(self._of_type)}]'

    def validate(self, value: Any):
        super().validate(value)
        for val in value:
            self._of_type.validate(val)

    def _from_string(self, value: str) -> list:
        values = value.replace(' ', '').lstrip('[').rstrip(']').split(',')
        if values == ['']:
            return []
        return values

    def convert(self, value: Any) -> list:
        converted_value = super().convert(value)
        if isinstance(converted_value, Iterable):
            res = []
            for val in converted_value:
                converted_val = self._of_type.convert(val)
                try:
                    self._of_type.validate(converted_val)
                except ValueError:
                    return value
                res.append(converted_val)
            return res
        return value

    def to_json(self, value: list) -> list:
        return [self._of_type.to_json(val) for val in value]

    def value_repr(self, value: list) -> str:
        if isinstance(self._of_type, (ObjectType, MappingType)) and value:
            return f'[...] ({len(value)} items)'
        return str(value)


class DictType(Type):
    _type = dict

    def __init__(self, keys: Type, values: Type):
        self._keys = keys
        self._values = values

    def __str__(self):
        return f'Dict[{str(self._keys)}, {str(self._values)}]'

    def validate(self, value: Any):
        super().validate(value)
        for key, val in value.items():
            self._keys.validate(key)
            self._values.validate(val)

    def _from_string(self, value: str) -> Union[dict, str]:
        try:
            return json.loads(value.replace("'", '"'))
        except json.JSONDecodeError:
            return value

    def convert(self, value: Any) -> Any:
        converted_value = super().convert(value)
        if isinstance(converted_value, dict):
            res = {}
            for key, val in converted_value.items():
                converted_key = self._keys.convert(key)
                converted_val = self._values.convert(val)
                try:
                    self._keys.validate(converted_key)
                    self._values.validate(converted_val)
                except ValueError:
                    return value
                res[converted_key] = converted_val
            return res
        return value

    def to_json(self, value: dict) -> dict:
        return {self._keys.to_json(k): self._values.to_json(v)
                for k, v in value.items()}

    def value_repr(self, value: dict) -> str:
        if isinstance(self._values, (ObjectType, MappingType)) and value:
            return f'{{...}} ({len(value)} items)'
        return str(value)


class DateTimeType(Type):
    _type = datetime
    _format = '%Y-%m-%dT%H:%M:%SZ'

    def _from_string(self, value: str) -> Union[datetime, str]:
        modified_value = re.sub(r'\..*Z', r'Z', value)
        try:
            return datetime.strptime(modified_value, self._format)
        except ValueError:
            return value

    def to_json(self, value: datetime) -> str:
        return datetime.strftime(value, self._format)


class ObjectType(Type):
    """
    Type representing some custom, user-defined class.
    It is assumed that objects of this class can be fully instantiated
    with a dict containing all the needed data.

    Example:

    .. code-block:: python

        >>> class Foo:
        ...     def __init__(self, a, b):
        ...         self.a = a
        ...         self.b = b
        ...
        ...     def __repr__(self):
        ...         return f'<Foo a={self.a}, b={self.b} />'

        >>> type_ = ObjectType(Foo)
        >>> type_.convert({'a': 123, 'b': 'bar'})
        <Foo a=123, b=bar />

    """

    def __init__(self, class_: TypingType[T]):
        self._type = class_

    def __str__(self):
        return str(self._type.__name__)

    def _from_string(self, value: str) -> T:
        raise NotImplementedError

    def convert(self, value: Any) -> T:
        if isinstance(value, dict):
            return self._type(**value)
        return value

    def to_json(self, value: T) -> Dict[str, Any]:
        return value.to_json()  # type: ignore[attr-defined]

    def value_repr(self, value: T) -> str:
        return f'<{self._type.__name__} ...>'


class MappingType(Type):
    """
    Type representing some kind of mapping. It is used to determine the
    correct type of the object at runtime, depending on the specified
    field.

    Example:

    .. code-block:: python

        >>> class Foo:
        ...     def __init__(self, a, type):
        ...         self.a = a
        ...         self.type = type
        ...
        ...     def __repr__(self):
        ...         return f'<Foo a={self.a} />'

        >>> class Bar:
        ...     def __init__(self, a, b, c, type):
        ...         self.abc = f'{a}{b}{c}'
        ...         self.type = type
        ...
        ...     def __repr__(self):
        ...         return f'<Bar abc={self.abc} />'

        >>> type_ = MappingType('type', {
        ...     'foo': ObjectType(Foo),
        ...     'bar': ObjectType(Bar),
        ... })

        >>> type_.convert({'type': 'foo', 'a': 123})
        <Foo a=123 />

        >>> type_.convert({'type': 'bar', 'a': 'a', 'b': 'b', 'c': 42})
        <Bar abc=ab42 />

        >>> foo = Foo(a=123, type='foo')
        >>> type_.validate(foo)

        >>> foo = Foo(a=123, type='bar')
        >>> type_.validate(foo)
        Traceback (most recent call last):
        ...
        ValueError

    """

    def __init__(self, mapping_key: str, mapping: Dict[str, ObjectType]):
        self._mapping_key = mapping_key
        self._mapping = mapping

    def __str__(self):
        types = ', '.join(type_ for type_ in self._mapping)
        return f'Union[{types}]'

    def convert(self, value: Any) -> Any:
        if isinstance(value, dict):
            key = value.get(to_camel_case(self._mapping_key), '')
            type_ = self._mapping.get(key)
            if not type_:
                return value
            return type_.convert(value)
        raise value

    def validate(self, value: Any):
        key = getattr(value, self._mapping_key, None)
        type_ = self._mapping.get(key)
        if not type_:
            raise ValueError
        type_.validate(value)

    def to_json(self, value: Any) -> Dict[str, Any]:
        return value.to_json()

    def value_repr(self, value: Any) -> str:
        key = getattr(value, self._mapping_key, None)
        type_ = self._mapping.get(key)
        if not type_:
            raise ValueError
        return type_.value_repr(value)


def new_type(type_: Type) -> Callable[[], Type]:
    return lambda: type_


NumericType = new_type(UnionType(IntegerType(), FloatType()))



"""
Module containing all classes and functions
responsible for dealing with snapshots.
"""
from __future__ import annotations

from typing import Any, Dict, ItemsView

from dsw_sdk.common.attributes import AttributesMixin


class Snapshot:
    """
    Snapshot of an object's state in a particular point of time.

    For now, it's basically just a dict.
    """

    def __init__(self, json_repr: Dict[str, Any]):
        self._json_repr = json_repr

    def __contains__(self, item):
        return self._json_repr.__contains__(item)

    def __getitem__(self, item):
        return self._json_repr.__getitem__(item)

    def items(self) -> ItemsView[str, Any]:
        """
        Mimics the built-in :meth:`dict.items` method.

        :return: Exactly the same result as a ``dict.items()`` would.
        """
        return self._json_repr.items()


class SnapshotDiff:
    """
    Class representing the differences between two snapshots (instances
    of :class:`Snapshot` class).

    Contains 3 categories:
        - what was modified (same keys, different values)
        - what was added (new keys)
        - what was deleted (old keys not present anymore)
    """

    def __init__(self):
        self.modified = {}
        self.added = {}
        self.removed = {}

    def __bool__(self):
        return bool({**self.modified, **self.added, **self.removed})

    def __contains__(self, item):
        return (
            item in self.modified
            or item in self.added
            or item in self.removed
        )

    def __len__(self):
        return len(self.modified) + len(self.added) + len(self.removed)


def make_snapshot(obj: AttributesMixin) -> Snapshot:
    """
    Creates a snapshot from :class:`AttributesMixin` instance.

    :param obj: instance of class :class:`AttributesMixin`

    :return: snapshot of ``obj`` state
    """
    return Snapshot(obj.to_json())


def snapshots_diff(old: Snapshot, new: Snapshot) -> SnapshotDiff:
    """
    Compares two snapshots (assuming the first one is older and the
    second one is newer), returning attributes in which they differ.

    :param old: former state of the object
    :param new: newer (current) state of the object

    :return: attributes that was added, changed or deleted
    """
    snapshot_diff = SnapshotDiff()

    for key, value in old.items():
        if key not in new:
            snapshot_diff.removed[key] = value
        elif value != new[key]:
            snapshot_diff.modified[key] = value

    for key, value in new.items():
        if key not in old:
            snapshot_diff.added[key] = value

    return snapshot_diff



"""
Templates high-level API.
"""
from __future__ import annotations

from typing import List

from dsw_sdk.high_level_api.common_api import API, RegistryAPIMixin
from dsw_sdk.high_level_api.models.templates.template import Template
from dsw_sdk.high_level_api.models.templates.template_asset import (
    TemplateAsset,
)
from dsw_sdk.high_level_api.models.templates.template_file import TemplateFile


class TemplateAPI(API, RegistryAPIMixin):
    """
    API for the Template entities.

    Supports all of the CRUD operations and pulling/updating templates.

    Example usage:

    .. code-block:: python

        api = TemplateAPI(...)

        # Get one template by ID
        template = api.get_template('dsw:template:1.0.0')

        # Get all templates for a given organization
        templates = api.get_templates(organization_id='dsw')

        # Get page number 1 (each page having 10 templates) of templates
        # containing the "foo" string, sorted by the UUID attribute in the
        # ascending order
        templates = api.get_templates(q='foo', page=1, size=10, start='id,asc')

        # Create template
        template = api.create_template(template_id='new_temp', ...)

        # Delete template
        api.delete_template(template.id)

        # Delete all templates for a given organization
        api.delete_templates(organization_id='dsw')

        # Delete list of templates
        api.delete_templates(ids=['dsw:template:1.0.0', 'dsw:temp:2.3.1'])

        # Pull specified templates from the Data Stewardship Registry
        api.pull_templates(['dsw:template:1.0.0', 'dsw:temp:2.3.1'])

        # Update specified templates (so that they have latest version
        # available)
        api.update_templates(['dsw:template:1.0.0', 'dsw:temp:2.3.1'])

        # Update all templates to the latest version available
        api.update_templates()
    """
    model_class = Template

    def _load_files(self, template_id: str) -> List[TemplateFile]:
        files_data = self._sdk.api.get_template_files(template_id).json()
        files = []
        for data in files_data:
            data['template_id'] = template_id
            files.append(TemplateFile(self._sdk, __update_attrs=data))
        return files

    def _load_assets(self, template_id: str) -> List[TemplateAsset]:
        assets_data = self._sdk.api.get_template_assets(template_id).json()
        assets = []
        for data in assets_data:
            asset_content = self._sdk.api.get_template_asset_content(
                template_id,
                data['uuid'],
            ).text
            data['template_id'] = template_id
            data['content'] = asset_content
            assets.append(TemplateAsset(self._sdk, __update_attrs=data))
        return assets

    def get_template(self, id_: str) -> Template:
        """
        Retrieves one template, identified by it's ID.
        Also loading all of it's related files and assets.

        :param id_: template identifier

        :return: object representing a template
        """
        template = self._get_one(self._sdk.api.get_template, id_)
        template._update_attrs(files=self._load_files(id_),
                               assets=self._load_assets(id_))
        return template

    def get_templates(
        self,
        organization_id: str = None,
        template_id: str = None,
        **query_params,
    ) -> List[Template]:
        """
        Retrieves list of templates.
        Also loading all related files and assets.

        Possibly identified by the organization to which they belong to. It's
        also possible to search based on the ``template_id`` (note that this
        is different from template's ``id`` attribute).

        :param organization_id: ID of the organization whose templates we
            want to fetch
        :param template_id: ``template_id`` attribute of a template
            (different from ``id`` attribute)
        :param query_params: optional query params: ``q``, ``size``, ``page``
            and ``sort``

        :return: list of objects, each representing a template
        """
        if organization_id is not None:
            query_params['organization_id'] = organization_id
        if template_id is not None:
            query_params['template_id'] = template_id
        templates = self._get_many_data(self._sdk.api.get_templates,
                                        'templates', **query_params)
        return [self.get_template(template['id']) for template in templates]

    def create_template(self, **kwargs) -> Template:
        """
        Creates a template with given data on the DSW server.

        :param kwargs: all the data needed for the template creation

        :return: object representing the new template
        """
        return self._create_new(**kwargs)

    def delete_template(self, id_: str):
        """
        Deletes a template from the DSW server.

        :param `id_`: ID of the template to delete
        """
        self._delete_one(self._sdk.api.delete_template, id_)

    def delete_templates(self, organization_id: str = None,
                         ids: List[str] = None):
        """
        Deletes multiple templates on the DSW server, identified either by
        organization or ID.

        :param organization_id: ID of the organization whose templates we
            want to delete
        :param ids: IDs of templates to delete
        """
        if not organization_id and not ids:
            raise ValueError('You must pass either `organization_id` or '
                             'template `ids` in order to delete templates. ')
        if organization_id:
            self._sdk.api.delete_templates(query_params={
                'organization_id': organization_id,
            })
        if ids:
            self._delete_many(self._sdk.api.delete_template, ids)

    def pull_templates(self, ids: List[str]):
        """
        Pulls given templates from the Data Stewardship Registry, so they
        become available in the DSW instance.

        :param ids: IDs of the templates you want to pull form the registry
        """
        self._pull(self._sdk.api.post_template_pull, 'Templates', ids)

    def update_templates(self, ids: List[str] = None):
        """
        Updates specified templates, pulling their latest available version.
        If no IDs are given, updates *all* templates on the DSW instance.

        :param ids: optional list of template IDs to update
        """
        self._update(self.get_template, self.get_templates,
                     self.pull_templates, ids)



# pylint: disable=R0901
"""
Module defining configuration of the whole library and implementing
the means of collecting the user config.

Configuration for each component is defined in separate class.
"""
from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from typing import Any, Dict, Optional, Tuple, Union

import yaml

from dsw_sdk.common.attributes import (
    Attribute,
    AttributesMixin,
    BoolAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import (
    DictType,
    IntegerType,
    NoneType,
    NumericType,
    StringType,
    TupleType,
    UnionType,
)


LOG_LEVELS = [logging.CRITICAL, logging.ERROR, logging.WARNING,
              logging.INFO, logging.DEBUG, logging.NOTSET]
# Extending the log levels with string representation of each level, so both
# e.g. `logging.INFO` and `20` are valid log level values.
LOG_LEVELS.extend([logging.getLevelName(level) for level in LOG_LEVELS])

MISSING_CONFIG_KEY_ERR = 'The config file `{}` does not contain the `{}` key.'

NUMERIC_PAIR_TYPE = TupleType(NumericType(), NumericType())
HEADERS_TYPE = DictType(StringType(), StringType())
TIMEOUT_TYPE = UnionType(NoneType(), NumericType(), NUMERIC_PAIR_TYPE)

Numeric = Union[int, float]
Headers = Dict[str, str]
Timeout = Union[None, Numeric, Tuple[Numeric, Numeric]]


class ComponentConfig(Mapping, AttributesMixin):
    """
    Base class for each component config.

    Implements all abstract methods of the :class:`collections.abc.Mapping`
    abstract class, so it can be used in following manner:

    .. code-block:: python

        >>> class Conf(ComponentConfig):
        ...     some_value = StringAttribute()

        >>> conf = Conf()
        >>> conf.some_value
        Traceback (most recent call last):
        ...
        dsw_sdk.common.attributes.AttributeNotSetError: ...
        >>> conf.some_value = 'foo'

        # Check if the value is set
        >>> 'some_value' in conf
        True

        # Get the item
        >>> conf.some_value
        'foo'
        >>> conf['some_value']
        'foo'

        # Get number of configured values
        >>> len(conf)
        1

        # Iterate over the config as dict
        >>> for value in conf.values(): pass
        >>> for key in conf.keys(): pass
        >>> for key, value in conf.items(): pass

        # Deconstruct the config to pass it as kwargs
        >>> def foo(**kwargs):
        ...     return kwargs['some_value']

        >>> foo(**conf)
        'foo'
    """
    def __contains__(self, item):
        return self.attrs().__contains__(item)

    def __getitem__(self, item):
        return self.attrs().__getitem__(item)

    def __iter__(self):
        return self.attrs().__iter__()

    def __len__(self):
        return self.attrs().__len__()


class HttpClientConfig(ComponentConfig):
    """
    Config for the HTTP client.
    """
    api_url: str = StringAttribute()
    email: str = StringAttribute()
    password: str = StringAttribute()
    enable_ssl: bool = BoolAttribute(default=True)
    auth_endpoint: str = StringAttribute(default='/tokens')
    headers: Headers = Attribute(HEADERS_TYPE, default={})
    default_timeout: Timeout = Attribute(TIMEOUT_TYPE,
                                         default=(6.05, 27))


class LoggerConfig(ComponentConfig):
    """
    Config for the Logger.
    """
    logger_name: str = StringAttribute(default='dsw_sdk')
    logger_level: str = Attribute(
        UnionType(IntegerType(), StringType()),
        default=logging.WARNING,
        choices=LOG_LEVELS,
    )
    logger_format: str = StringAttribute(
        default='[%(asctime)s] - %(name)s | %(levelname)s | %(message)s'
    )


class Config:
    """
    This class serves 2 purposes.

    It contains all the other "partial" configs. E.g.
    config objects for HTTP client or logger.

    It also collects user-defined configuration from env variables (prefixed
    by ``DSW_SDK``) and from file (YAML containing ``dsw_sdk`` section) passed
    in ``conf_file`` keyword argument. Then it merges these with all other
    values passed as keyword arguments, in following order (first takes
    precedence over the others):

        1) keyword arguments
        2) environment variables
        3) file config

    Example file config:

    .. code-block:: yaml

        dsw_sdk:
          enable_ssl: false
          auth_endpoint: '/auth'
          headers:
            'X-CUSTOM-HEADER': 'Custom value'
          default_timeout:
            - 6
            - 120

    """
    _FILE_SECTION = 'dsw_sdk'
    _ENV_PREFIX = 'DSW_SDK_'

    def __init__(self, **obj_config):
        """
        :param obj_config: arbitrary config values passed as keyword arguments;
                           if path is passed in ``conf_file``, it tries to load
                           the config values from a file
        """
        conf_file = obj_config.pop('conf_file', None)
        file_config = self._init_file_config(conf_file)
        env_config = self._init_env_config()
        conf = {**file_config, **env_config, **obj_config}

        self.http_client = HttpClientConfig(**conf)
        self.logger = LoggerConfig(**conf)

    @classmethod
    def _init_file_config(cls, conf_file: Optional[str]) -> Dict[str, Any]:
        if not conf_file:
            return {}

        with open(conf_file, 'r', encoding='utf-8') as file:
            file_config = yaml.safe_load(file)
            if cls._FILE_SECTION not in file_config:
                raise KeyError(
                    MISSING_CONFIG_KEY_ERR.format(conf_file, cls._FILE_SECTION)
                )
            return file_config.get(cls._FILE_SECTION, {})

    @classmethod
    def _init_env_config(cls) -> Dict[str, Any]:
        env_config = {}

        for k, v in os.environ.items():  # pylint: disable=C0103
            if not k.upper().startswith(cls._ENV_PREFIX):
                continue
            key = k.lstrip(cls._ENV_PREFIX).lower()
            env_config.update({key: v})

        return env_config



from dsw_sdk.common.attributes import (
    DateTimeAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.high_level_api.dto.questionnaire import QuestionnaireSimple
from dsw_sdk.high_level_api.dto.template import TemplateSimpleDTO
from dsw_sdk.high_level_api.models.model import Model


DONE_DOCUMENT_STATE = 'DoneDocumentState'
ERROR_DOCUMENT_STATE = 'ErrorDocumentState'
IN_PROGRESS_DOCUMENT_STATE = 'InProgressDocumentState'
QUEUED_DOCUMENT_STATE = 'QueuedDocumentState'
DOCUMENT_STATES = (
    DONE_DOCUMENT_STATE,
    ERROR_DOCUMENT_STATE,
    IN_PROGRESS_DOCUMENT_STATE,
    QUEUED_DOCUMENT_STATE,
)


class Document(Model):
    created_at = DateTimeAttribute()
    creator_uuid = StringAttribute(nullable=True)
    format_uuid = StringAttribute()
    name = StringAttribute()
    questionnaire = ObjectAttribute(QuestionnaireSimple, nullable=True)
    questionnaire_event_uuid = StringAttribute(nullable=True)
    state = StringAttribute(choices=DOCUMENT_STATES)
    template = ObjectAttribute(TemplateSimpleDTO)

    def _create(self):
        raise NotImplementedError('Cannot create documents')

    def _update(self):
        raise NotImplementedError('Cannot update documents')

    def _delete(self):
        raise NotImplementedError('Cannot delete documents')



from dsw_sdk.common.attributes import (
    AttributesMixin,
    BoolAttribute, DictAttribute, ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import ObjectType, StringType
from dsw_sdk.high_level_api.dto.questionnaire import (
    QUESTIONNAIRE_SHARING,
    QUESTIONNAIRE_VISIBILITIES,
)


CUSTOM_QUESTIONNAIRE_CREATION = 'CustomQuestionnaireCreation'
TEMPLATE_QUESTIONNAIRE_CREATION = 'TemplateQuestionnaireCreation'
TEMPLATE_AND_CUSTOM_QUESTIONNAIRE_CREATION = (
    'TemplateAndCustomQuestionnaireCreation'
)
QUESTIONNAIRE_CREATION = (
    CUSTOM_QUESTIONNAIRE_CREATION,
    TEMPLATE_QUESTIONNAIRE_CREATION,
    TEMPLATE_AND_CUSTOM_QUESTIONNAIRE_CREATION,
)


class PackagePattern(AttributesMixin):
    km_id = StringAttribute(nullable=True)
    max_version = StringAttribute(nullable=True)
    min_version = StringAttribute(nullable=True)
    org_id = StringAttribute(nullable=True)


class AppConfigKnowledgeModelPublic(AttributesMixin):
    enabled = BoolAttribute()
    packages = ListAttribute(ObjectType(PackagePattern))


class AppConfigKnowledgeModel(AttributesMixin):
    public = ObjectAttribute(AppConfigKnowledgeModelPublic)


class AppConfigOrganization(AttributesMixin):
    affiliations = ListAttribute(StringType())
    description = StringAttribute()
    name = StringAttribute()
    organization_id = StringAttribute()


class SimpleFeature(AttributesMixin):
    enabled = BoolAttribute()


class AppConfigAuthInternal(AttributesMixin):
    registration = ObjectAttribute(SimpleFeature)


class AppConfigAuthExternalServiceParameter(AttributesMixin):
    name = StringAttribute()
    value = StringAttribute()


class AppConfigAuthExternalServiceStyle(AttributesMixin):
    background = StringAttribute(nullable=True)
    color = StringAttribute(nullable=True)
    icon = StringAttribute(nullable=True)


class AppConfigAuthExternalService(AttributesMixin):
    client_id = StringAttribute()
    client_secret = StringAttribute()
    id = StringAttribute()
    name = StringAttribute()
    parameteres = ListAttribute(
        ObjectType(AppConfigAuthExternalServiceParameter)
    )
    style = ObjectAttribute(AppConfigAuthExternalServiceStyle, nullable=True)
    url = StringAttribute()


class AppConfigAuthExternal(AttributesMixin):
    services = ListAttribute(ObjectType(AppConfigAuthExternalService))


class AppConfigAuth(AttributesMixin):
    default_role = StringAttribute()
    internal = ObjectAttribute(AppConfigAuthInternal)
    external = ObjectAttribute(AppConfigAuthExternal)


class AppConfigPrivacyAndSupport(AttributesMixin):
    privacy_url = StringAttribute(nullable=True)
    support_email = StringAttribute(nullable=True)
    support_repository_name = StringAttribute(nullable=True)
    support_repository_url = StringAttribute(nullable=True)
    terms_of_service_url = StringAttribute(nullable=True)


class AppConfigDashboardWidgets(AttributesMixin):
    admin = ListAttribute(StringType())
    data_steward = ListAttribute(StringType())
    researcher = ListAttribute(StringType())


class AppConfigDashboard(AttributesMixin):
    welcome_info = StringAttribute(nullable=True)
    welcome_warning = StringAttribute(nullable=True)
    widgets = ObjectAttribute(AppConfigDashboardWidgets, nullable=True)


class AppConfigLookAndFeelCustomMenuLink(AttributesMixin):
    icon = StringAttribute()
    new_window = BoolAttribute()
    title = StringAttribute()
    url = StringAttribute()


class AppConfigLookAndFeel(AttributesMixin):
    app_title = StringAttribute(nullable=True)
    app_title_short = StringAttribute(nullable=True)
    custom_menu_links = ListAttribute(
        ObjectType(AppConfigLookAndFeelCustomMenuLink)
    )
    login_info = StringAttribute(nullable=True)


class AppConfigRegistry(AttributesMixin):
    enabled = BoolAttribute()
    token = StringAttribute()


class AppConfigQuestionnaireVisibility(AttributesMixin):
    default_value = StringAttribute(choices=QUESTIONNAIRE_VISIBILITIES)
    enabled = BoolAttribute()


class AppConfigQuestionnaireSharing(AttributesMixin):
    anonymous_enabled = BoolAttribute()
    default_value = StringAttribute(choices=QUESTIONNAIRE_SHARING)
    enabled = BoolAttribute()


class AppConfigQuestionnaireFeedback(AttributesMixin):
    enabled = BoolAttribute()
    owner = StringAttribute()
    repo = StringAttribute()
    token = StringAttribute()


class AppConfigQuestionnaire(AttributesMixin):
    feedback = ObjectAttribute(AppConfigQuestionnaireFeedback)
    questionnaire_creation = StringAttribute(choices=QUESTIONNAIRE_CREATION)
    questionnaire_sharing = ObjectAttribute(AppConfigQuestionnaireSharing)
    questionnaire_visibility = ObjectAttribute(
        AppConfigQuestionnaireVisibility
    )
    summary_report = ObjectAttribute(SimpleFeature)


class AppConfigTemplate(AttributesMixin):
    recommended_template_id = StringAttribute(nullable=True)


class AppConfigSubmissionServiceSupportedFormat(AttributesMixin):
    format_uuid = StringAttribute()
    template_id = StringAttribute()


class AppConfigSubmissionServiceRequestMultipart(AttributesMixin):
    enabled = BoolAttribute()
    file_name = StringAttribute()


class AppConfigSubmissionServiceRequest(AttributesMixin):
    headers = DictAttribute(StringType(), StringType())
    method = StringAttribute()
    multipart = ObjectAttribute(AppConfigSubmissionServiceRequestMultipart)
    url = StringAttribute()


class AppConfigSubmissionService(AttributesMixin):
    description = StringAttribute()
    id = StringAttribute()
    name = StringAttribute()
    props = ListAttribute(StringType())
    request = ObjectAttribute(AppConfigSubmissionServiceRequest)
    supported_formats = ListAttribute(
        ObjectType(AppConfigSubmissionServiceSupportedFormat)
    )


class AppConfigSubmission(AttributesMixin):
    enabled = BoolAttribute()
    services = ListAttribute(ObjectType(AppConfigSubmissionService))


class AppConfigChangeDTO(AttributesMixin):
    authentication = ObjectAttribute(AppConfigAuth)
    dashboard = ObjectAttribute(AppConfigDashboard)
    knowledge_model = ObjectAttribute(AppConfigKnowledgeModel)
    look_and_feel = ObjectAttribute(AppConfigLookAndFeel)
    organization = ObjectAttribute(AppConfigOrganization)
    privacy_and_support = ObjectAttribute(AppConfigPrivacyAndSupport)
    questionnaire = ObjectAttribute(AppConfigQuestionnaire)
    registry = ObjectAttribute(AppConfigRegistry)
    submission = ObjectAttribute(AppConfigSubmission)
    template = ObjectAttribute(AppConfigTemplate)



# pylint: disable-all

from __future__ import annotations

from typing import Any, Dict, Optional

from dsw_sdk.common.utils import to_camel_case
from dsw_sdk.http_client.interface import HttpClient, HttpResponse


class LowLevelAPI:
    """
    Low-level API mirroring 1:1 the Data Stewardship Wizard API. It contains
    one method for each combination of HTTP method and API endpoint.

    If the endpoint accepts query parameters or body, the method accept
    these as well. Keys in both query params and body are converted to
    `camelCase`, so you can pass them in `snake_case` if you want.

    Note that this class is *generated* by a script, not written by hand.
    """
    def __init__(self, http_client: HttpClient):
        """
        :param http_client: Some instance of the :class:`~interface.HttpClient`
                            interface.
        """
        self._http_client = http_client

    def _camelize_dict_keys(self, data: Any) -> Any:
        if isinstance(data, dict):
            return {to_camel_case(k): self._camelize_dict_keys(v)
                    for k, v in data.items()}
        elif isinstance(data, list):
            return [self._camelize_dict_keys(val) for val in data]
        else:
            return data
    
    def post_questionnaire_migrations(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            targetTagUuids: array
            targetPackageId: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/migrations', body=body, **kwargs)
    
    def post_package_pull(self, pkg_id: str, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/packages/{pkg_id}/pull', **kwargs)
    
    def post_branch_migrations_current_conflict(self, b_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            originalEventUuid: None
            action: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/branches/{b_uuid}/migrations/current/conflict', body=body, **kwargs)
    
    def post_templates(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            readme: string
            templateId: string
            formats: array
            metamodelVersion: integer
            name: string
            version: string
            license: string
            organizationId: string
            description: string
            allowedPackages: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/templates', body=body, **kwargs)
    
    def get_templates(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            organizationId [optional]: string
            templateId [optional]: string
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/templates', params=query_params, **kwargs)
    
    def delete_templates(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            organizationId [optional]: string
            templateId [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.delete(f'/templates', params=query_params, **kwargs)
    
    def get_questionnaire(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/questionnaires/{qtn_uuid}', **kwargs)
    
    def delete_questionnaire(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/questionnaires/{qtn_uuid}', **kwargs)
    
    def put_questionnaire(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            isTemplate: boolean
            visibility: None
            name: string
            permissions: array
            sharing: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/questionnaires/{qtn_uuid}', body=body, **kwargs)
    
    def delete_document(self, doc_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/documents/{doc_uuid}', **kwargs)
    
    def post_questionnaires_from_template(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            name: string
            questionnaireUuid: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/questionnaires/from-template', body=body, **kwargs)
    
    def get_feedback(self, f_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/feedbacks/{f_uuid}', **kwargs)
    
    def post_templates_bundle(self, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/templates/bundle', **kwargs)
    
    def get_questionnaire_documents_preview(self, qtn_uuid: str, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            Authorization [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/questionnaires/{qtn_uuid}/documents/preview', params=query_params, **kwargs)
    
    def put_questionnaire_content(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            events: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/questionnaires/{qtn_uuid}/content', body=body, **kwargs)
    
    def get_templates_all(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            organizationId [optional]: string
            templateId [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/templates/all', params=query_params, **kwargs)
    
    def get_questionnaire_migrations_current(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/questionnaires/{qtn_uuid}/migrations/current', **kwargs)
    
    def delete_questionnaire_migrations_current(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/questionnaires/{qtn_uuid}/migrations/current', **kwargs)
    
    def put_questionnaire_migrations_current(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            resolvedQuestionUuids: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/questionnaires/{qtn_uuid}/migrations/current', body=body, **kwargs)
    
    def post_template_files(self, template_id: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            content: string
            fileName: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/templates/{template_id}/files', body=body, **kwargs)
    
    def get_template_files(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}/files', **kwargs)
    
    def post_registry_confirmation(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            hash: string
            organizationId: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/registry/confirmation', body=body, **kwargs)
    
    def post_users(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            email: string
            lastName: string
            password: string
            firstName: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/users', body=body, **kwargs)
    
    def get_users(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            q [optional]: string
            role [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/users', params=query_params, **kwargs)
    
    def post_branch_migrations_current(self, b_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            targetTagUuids: array
            targetPackageId: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/branches/{b_uuid}/migrations/current', body=body, **kwargs)
    
    def get_branch_migrations_current(self, b_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/branches/{b_uuid}/migrations/current', **kwargs)
    
    def delete_branch_migrations_current(self, b_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/branches/{b_uuid}/migrations/current', **kwargs)
    
    def get_package_bundle(self, pkg_id: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/packages/{pkg_id}/bundle', **kwargs)
    
    def get_questionnaire_report(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/questionnaires/{qtn_uuid}/report', **kwargs)
    
    def get_template_asset(self, template_id: str, asset_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}/assets/{asset_uuid}', **kwargs)
    
    def delete_template_asset(self, template_id: str, asset_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/templates/{template_id}/assets/{asset_uuid}', **kwargs)
    
    def post_questionnaire_clone(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/clone', **kwargs)
    
    def delete_questionnaire_version(self, qtn_uuid: str, v_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/questionnaires/{qtn_uuid}/versions/{v_uuid}', **kwargs)
    
    def put_questionnaire_version(self, qtn_uuid: str, v_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            eventUuid: None
            name: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/questionnaires/{qtn_uuid}/versions/{v_uuid}', body=body, **kwargs)
    
    def post_action_keys(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            email: string
            type: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/action-keys', body=body, **kwargs)
    
    def get_users_current(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/users/current', **kwargs)
    
    def put_users_current(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            email: string
            lastName: string
            firstName: string
            submissionProps: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/users/current', body=body, **kwargs)
    
    def get_configs_bootstrap(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/configs/bootstrap', **kwargs)
    
    def post_questionnaires(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            tagUuids: array
            packageId: string
            visibility: None
            name: string
            sharing: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/questionnaires', body=body, **kwargs)
    
    def get_questionnaires(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            q [optional]: string
            isTemplate [optional]: boolean
            userUuids [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/questionnaires', params=query_params, **kwargs)
    
    def get_auth(self, id: str, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            clientUrl [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/auth/{id}', params=query_params, **kwargs)
    
    def post_document_submissions(self, doc_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            serviceId: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/documents/{doc_uuid}/submissions', body=body, **kwargs)
    
    def get_document_submissions(self, doc_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/documents/{doc_uuid}/submissions', **kwargs)
    
    def post_registry_signup(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            email: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/registry/signup', body=body, **kwargs)
    
    def post_packages(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        return self._http_client.post(f'/packages', **kwargs)
    
    def get_packages(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            organizationId [optional]: string
            kmId [optional]: string
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/packages', params=query_params, **kwargs)
    
    def delete_packages(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            organizationId [optional]: string
            kmId [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.delete(f'/packages', params=query_params, **kwargs)
    
    def post_questionnaire_revert(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            eventUuid: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/revert', body=body, **kwargs)
    
    def put_user_state(self, u_uuid: str, body: Dict[str, Any], query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        body:
            active: boolean
        
        query_params:
            hash [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/users/{u_uuid}/state', body=body, params=query_params, **kwargs)
    
    def put_user_password(self, u_uuid: str, body: Dict[str, Any], query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        body:
            password: string
        
        query_params:
            hash [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/users/{u_uuid}/password', body=body, params=query_params, **kwargs)
    
    def post_template_assets(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/templates/{template_id}/assets', **kwargs)
    
    def get_template_assets(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}/assets', **kwargs)
    
    def get_user(self, u_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/users/{u_uuid}', **kwargs)
    
    def delete_user(self, u_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/users/{u_uuid}', **kwargs)
    
    def put_user(self, u_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            email: string
            lastName: string
            active: boolean
            role: string
            firstName: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/users/{u_uuid}', body=body, **kwargs)
    
    def post_questionnaire_revert_preview(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            eventUuid: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/revert/preview', body=body, **kwargs)
    
    def get_document_download(self, doc_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/documents/{doc_uuid}/download', **kwargs)
    
    def get_templates_suggestions(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            pkgId [optional]: string
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/templates/suggestions', params=query_params, **kwargs)
    
    def post_tokens(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            email: string
            password: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/tokens', body=body, **kwargs)
    
    def post_caches_knowledge_model(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            tagUuids: array
            events: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/caches/knowledge-model', body=body, **kwargs)
    
    def get_document_available_submission_services(self, doc_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/documents/{doc_uuid}/available-submission-services', **kwargs)
    
    def post_typehints(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            questionUuid: None
            q: string
            events: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/typehints', body=body, **kwargs)
    
    def post_admin_operations_executions(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            operationName: string
            parameters: array
            sectionName: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/admin/operations/executions', body=body, **kwargs)
    
    def get_(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/', **kwargs)
    
    def get_questionnaire_documents(self, qtn_uuid: str, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/questionnaires/{qtn_uuid}/documents', params=query_params, **kwargs)
    
    def post_questionnaire_versions(self, qtn_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            eventUuid: None
            name: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/versions', body=body, **kwargs)
    
    def get_questionnaire_versions(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/questionnaires/{qtn_uuid}/versions', **kwargs)
    
    def post_template_pull(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/templates/{template_id}/pull', **kwargs)
    
    def get_auth_callback(self, id: str, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            clientUrl [optional]: string
            error [optional]: string
            code [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/auth/{id}/callback', params=query_params, **kwargs)
    
    def post_packages_bundle(self, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/packages/bundle', **kwargs)
    
    def post_feedbacks(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            packageId: string
            questionUuid: None
            content: string
            title: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/feedbacks', body=body, **kwargs)
    
    def get_feedbacks(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            packageId [optional]: string
            questionUuid [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/feedbacks', params=query_params, **kwargs)
    
    def post_branches(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            kmId: string
            name: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/branches', body=body, **kwargs)
    
    def get_branches(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/branches', params=query_params, **kwargs)
    
    def get_users_suggestions(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/users/suggestions', params=query_params, **kwargs)
    
    def get_book_reference(self, br_short_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/book-references/{br_short_uuid}', **kwargs)
    
    def get_feedbacks_synchronization(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/feedbacks/synchronization', **kwargs)
    
    def get_template_file(self, template_id: str, file_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}/files/{file_uuid}', **kwargs)
    
    def delete_template_file(self, template_id: str, file_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/templates/{template_id}/files/{file_uuid}', **kwargs)
    
    def put_template_file(self, template_id: str, file_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            content: string
            fileName: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/templates/{template_id}/files/{file_uuid}', body=body, **kwargs)
    
    def get_template(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}', **kwargs)
    
    def delete_template(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/templates/{template_id}', **kwargs)
    
    def put_template(self, template_id: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            readme: string
            templateId: string
            formats: array
            metamodelVersion: integer
            name: string
            version: string
            license: string
            organizationId: string
            description: string
            allowedPackages: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/templates/{template_id}', body=body, **kwargs)
    
    def delete_caches(self, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/caches', **kwargs)
    
    def get_configs_app(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/configs/app', **kwargs)
    
    def put_configs_app(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            privacyAndSupport: None
            authentication: None
            lookAndFeel: None
            dashboard: None
            questionnaire: None
            registry: None
            knowledgeModel: None
            submission: None
            template: None
            organization: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/configs/app', body=body, **kwargs)
    
    def put_users_current_password(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            password: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/users/current/password', body=body, **kwargs)
    
    def get_branch(self, b_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/branches/{b_uuid}', **kwargs)
    
    def delete_branch(self, b_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/branches/{b_uuid}', **kwargs)
    
    def put_branch(self, b_uuid: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            kmId: string
            name: string
            events: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/branches/{b_uuid}', body=body, **kwargs)
    
    def post_documents(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            templateId: string
            name: string
            questionnaireUuid: None
            formatUuid: None
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/documents', body=body, **kwargs)
    
    def get_documents(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            questionnaireUuid [optional]: string
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/documents', params=query_params, **kwargs)
    
    def get_documents_housekeeping(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/documents/housekeeping', **kwargs)
    
    def get_template_bundle(self, template_id: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}/bundle', **kwargs)
    
    def get_template_asset_content(self, template_id: str, asset_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/templates/{template_id}/assets/{asset_uuid}/content', **kwargs)
    
    def get_packages_suggestions(self, query_params: Optional[Dict[str, Any]] = None, **kwargs) -> HttpResponse:
        """
        query_params:
            q [optional]: string
            page [optional]: integer
            size [optional]: integer
            sort [optional]: string
        
        """
        query_params = self._camelize_dict_keys(query_params)
        return self._http_client.get(f'/packages/suggestions', params=query_params, **kwargs)
    
    def put_branch_version(self, b_uuid: str, version: str, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            readme: string
            license: string
            description: string
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.put(f'/branches/{b_uuid}/versions/{version}', body=body, **kwargs)
    
    def post_questionnaire_squash(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/squash', **kwargs)
    
    def get_package(self, pkg_id: str, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/packages/{pkg_id}', **kwargs)
    
    def delete_package(self, pkg_id: str, **kwargs) -> HttpResponse:
        return self._http_client.delete(f'/packages/{pkg_id}', **kwargs)
    
    def post_knowledge_models_preview(self, body: Dict[str, Any], **kwargs) -> HttpResponse:
        """
        body:
            tagUuids: array
            events: array
        
        """
        body = self._camelize_dict_keys(body)
        return self._http_client.post(f'/knowledge-models/preview', body=body, **kwargs)
    
    def post_questionnaire_migrations_current_completion(self, qtn_uuid: str, **kwargs) -> HttpResponse:
        return self._http_client.post(f'/questionnaires/{qtn_uuid}/migrations/current/completion', **kwargs)
    
    def get_admin_operations(self, **kwargs) -> HttpResponse:
        return self._http_client.get(f'/admin/operations', **kwargs)



"""
Questionnaires high-level API.
"""
from __future__ import annotations

from typing import List

from dsw_sdk.high_level_api.common_api import API
from dsw_sdk.high_level_api.models.questionnaire import Questionnaire


class QuestionnaireAPI(API):
    """
    API for the Questionnaire entities.

    For now, there are just two methods for getting one or many questionnaires
    (as others were not yet required, but might be implemented in the future).

    Example usage:

    .. code-block:: python

        api = QuestionnaireAPI(...)

        # Get one questionnaire by UUID
        q = api.get_questionnaire('some-uuid-1234')

        # Get page number 1 (each page having 10 questionnaires) of
        # questionnaires containing the "foo" string, sorted by the UUID
        # attribute in the ascending order
        qs = api.get_questionnaires(q='foo', page=1, size=10, sort='uuid,asc')
    """
    model_class = Questionnaire

    def get_questionnaire(self, uuid: str) -> Questionnaire:
        """
        Retrieves one questionnaire, identified by it's UUID.
        Also loading all of it's related documents.

        :param uuid: universally unique identifier of the questionnaire

        :return: object representing a questionnaire
        """
        questionnaire = self._get_one(self._sdk.api.get_questionnaire, uuid)
        documents_page = self._sdk.api.get_questionnaire_documents(uuid).json()
        questionnaire._update_attrs(
            package_id=questionnaire.package.id,
            documents=documents_page['_embedded']['documents']
        )
        return questionnaire

    def get_questionnaires(self, **query_params) -> List[Questionnaire]:
        """
        Retrieves list of questionnaires.
        Also loading all related documents.

        :param query_params: optional query params ``q``, ``size``, ``page``
            and ``sort``

        :return: list of objects, each representing a questionnaire
        """
        questionnaires = self._get_many_data(self._sdk.api.get_questionnaires,
                                             'questionnaires', **query_params)
        return [self.get_questionnaire(questionnaire['uuid'])
                for questionnaire in questionnaires]

    def create_questionnaire(self, **kwargs) -> Questionnaire:
        """
        Create a new questionnaire (custom).

        :param kwargs: all the data needed for the questionnaire creation

        :return: object representing the newly created questionnaire
        """
        return self._create_new(**kwargs)

    def create_questionnaire_from_template(
            self, name: str, questionnaire_uuid: str
    ) -> Questionnaire:
        """
        Create a new questionnaire from a template

        :param name: name of the new questionnaire
        :param questionnaire_uuid: uuid of the template questionnaire

        :return: object representing the newly created questionnaire
        """
        questionnaire = self.model_class(self._sdk)
        questionnaire.create_from_template(
            name=name,
            questionnaire_uuid=questionnaire_uuid,
        )
        return questionnaire



from dsw_sdk.common.attributes import DateTimeAttribute, ObjectAttribute
from dsw_sdk.high_level_api.dto.app_config import (
    AppConfigAuth,
    AppConfigChangeDTO,
    AppConfigDashboard,
    AppConfigKnowledgeModel,
    AppConfigLookAndFeel,
    AppConfigOrganization,
    AppConfigPrivacyAndSupport,
    AppConfigQuestionnaire,
    AppConfigRegistry,
    AppConfigSubmission,
    AppConfigTemplate,
)
from dsw_sdk.high_level_api.models.model import Model


class AppConfig(Model):
    authentication = ObjectAttribute(AppConfigAuth)
    created_at = DateTimeAttribute()
    dashboard = ObjectAttribute(AppConfigDashboard)
    knowledge_model = ObjectAttribute(AppConfigKnowledgeModel)
    look_and_feel = ObjectAttribute(AppConfigLookAndFeel)
    organization = ObjectAttribute(AppConfigOrganization)
    privacy_and_support = ObjectAttribute(AppConfigPrivacyAndSupport)
    questionnaire = ObjectAttribute(AppConfigQuestionnaire)
    registry = ObjectAttribute(AppConfigRegistry)
    submission = ObjectAttribute(AppConfigSubmission)
    template = ObjectAttribute(AppConfigTemplate)
    updated_at = DateTimeAttribute()

    def _create(self):
        raise NotImplementedError('Cannot create config')

    def _update(self):
        dto = AppConfigChangeDTO(**self.attrs())
        dto.validate()
        data = self._sdk.api.put_configs_app(body=dto.to_json()).json()
        self._update_attrs(**data)

    def _delete(self):
        raise NotImplementedError('Cannot delete config')



# pylint: disable=E1133

from datetime import datetime
from typing import Any, List, Optional

from dsw_sdk.common.attributes import (
    Alias,
    DateTimeAttribute,
    IntegerAttribute,
    ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.snapshot import SnapshotDiff, make_snapshot, snapshots_diff
from dsw_sdk.common.types import ObjectType, StringType
from dsw_sdk.common.utils import truncate_long_string
from dsw_sdk.high_level_api.dto.common import (
    OrganizationSimple,
    PackageSimpleDTO,
)
from dsw_sdk.high_level_api.dto.template import (
    TEMPLATE_STATES,
    TemplateAllowedPackage,
    TemplateChangeDTO, TemplateFormat,
)
from dsw_sdk.high_level_api.models.model import (
    ListOfModelsAttribute,
    Model,
)
from dsw_sdk.high_level_api.models.templates.template_asset import (
    TemplateAsset,
)
from dsw_sdk.high_level_api.models.templates.template_file import TemplateFile


CREATE_VALIDATE_ERR = ('Invalid `{}` attribute: `template_id` of a {} '
                       'cannot be set when creating new template')
UPDATE_VALIDATE_ERR = ('Invalid `{}` attribute: `template_id` of a {} must be '
                       'same as `uuid` attribute of corresponding template')


class Template(Model):
    allowed_packages: List[TemplateAllowedPackage] = ListAttribute(
        ObjectType(TemplateAllowedPackage),
    )
    created_at: datetime = DateTimeAttribute(read_only=True)
    description: str = StringAttribute()
    formats: List[TemplateFormat] = ListAttribute(ObjectType(TemplateFormat))
    id: str = Alias('uuid')
    license: str = StringAttribute()
    metamodel_version: int = IntegerAttribute()
    name: str = StringAttribute()
    organization: Optional[OrganizationSimple] = ObjectAttribute(
        OrganizationSimple,
        nullable=True,
        read_only=True,
    )
    organization_id: str = StringAttribute()
    readme: str = StringAttribute()
    recommended_package_id: Optional[str] = StringAttribute(nullable=True)
    registry_link: Optional[str] = StringAttribute(
        nullable=True,
        read_only=True,
    )
    remote_latest_version: Optional[str] = StringAttribute(
        nullable=True,
        read_only=True,
    )
    state: str = StringAttribute(choices=TEMPLATE_STATES, read_only=True)
    template_id: str = StringAttribute()
    usable_packages: List[PackageSimpleDTO] = ListAttribute(
        ObjectType(PackageSimpleDTO),
        read_only=True,
    )
    version: str = StringAttribute()
    versions: List[str] = ListAttribute(StringType(), read_only=True)

    assets: List[TemplateAsset] = ListOfModelsAttribute(TemplateAsset,
                                                        default=[])
    files: List[TemplateFile] = ListOfModelsAttribute(TemplateFile,
                                                      default=[])

    def _attr_to_str(self, name: str, value: Any) -> str:
        # Readme is usually quite long, so we display only the beginning
        if name == 'readme':
            return truncate_long_string(value, 50)
        return super()._attr_to_str(name, value)

    def _create_validate(self):
        for file in self.files:
            if file.template_id is not None:
                raise ValueError(CREATE_VALIDATE_ERR.format('files', 'file'))
        for asset in self.assets:
            if asset.template_id is not None:
                raise ValueError(CREATE_VALIDATE_ERR.format('assets', 'asset'))

    def _update_validate(self):
        for file in self.files:
            if (
                file.template_id is not None
                and file.template_id != self.uuid
            ):
                raise ValueError(UPDATE_VALIDATE_ERR.format('files', 'file'))
        for asset in self.assets:
            if (
                asset.template_id is not None
                and asset.template_id != self.uuid
            ):
                raise ValueError(UPDATE_VALIDATE_ERR.format('assets', 'asset'))

    def _save_template_files(self, diff: SnapshotDiff = SnapshotDiff()):
        modified_files_uuids = [file['uuid'] for file in
                                diff.modified.get('files', [])]
        for file in self.files:
            if file.uuid in modified_files_uuids:
                modified_files_uuids.remove(file.uuid)
            file.template_id = self.uuid
            # If it is a new file, it gets created. If it is some old file,
            # that was just modified, it gets updated. If no there was no
            # change, `file.save()` won't do anything.
            file.save()

        for uuid_to_remove in modified_files_uuids:
            self._sdk.api.delete_template_file(self.uuid, uuid_to_remove)

    def _save_asset_files(self, diff: SnapshotDiff = SnapshotDiff()):
        modified_assets_uuids = [asset['uuid'] for asset in
                                 diff.modified.get('assets', [])]

        for asset in self.assets:
            if asset.uuid in modified_assets_uuids:
                modified_assets_uuids.remove(asset.uuid)
            asset.template_id = self.uuid
            asset.save()

        for uuid_to_remove in modified_assets_uuids:
            self._sdk.api.delete_template_asset(self.uuid, uuid_to_remove)

    def _create(self):
        self._create_validate()
        dto = TemplateChangeDTO(**self.attrs())
        dto.validate()
        data = self._sdk.api.post_templates(body=dto.to_json()).json()
        # We must pop the `files` and `assets`, because they are not yet
        # created on the server, so there are only empty lists in `data`.
        data.pop('files', None)
        data.pop('assets', None)
        self._update_attrs(**data)
        self._save_template_files()
        self._save_asset_files()

        data = self._sdk.api.get_template(self.id).json()
        self._update_attrs(**data)

    def _update(self):
        self._update_validate()
        diff = snapshots_diff(self._snapshot, make_snapshot(self))

        if 'files' in diff:
            self._save_template_files(diff)
        if 'assets' in diff:
            self._save_asset_files(diff)
        if len(diff) > 2 or ('files' not in diff and 'assets' not in diff):
            dto = TemplateChangeDTO(**self.attrs())
            dto.validate()
            data = self._sdk.api.put_template(
                self.uuid,
                body=dto.to_json(),
            ).json()
            self._update_attrs(**data)

    def _delete(self):
        self._sdk.api.delete_template(self.uuid)



"""
Module with classes procuring the whole authentication process.
"""
from __future__ import annotations

from datetime import datetime, timedelta
from typing import Dict

import jwt
import requests
from requests.auth import AuthBase

from dsw_sdk.http_client.interface import UnexpectedAuthError


AUTH_ERROR = 'There is no token in authentication response. ' \
             'Please, contact the API maintainer.'
INVALID_TOKEN_ERROR = 'Authentication token is invalid. ' \
                      'Please, contact the API maintainer.'


class TokenRetrievalStrategy:
    """
    Abstract class for strategy on how to retrieve auth token from the server.

    Every subclass has to implement the :meth:`get_new_token` method.
    """
    def __init__(self, auth_url: str, credentials: Dict[str, str]):
        """
        :param auth_url: URL of the authentication endpoint
        :param credentials: dict containing keys ``email`` and ``password``
        """
        self._url = auth_url
        self._credentials = credentials

    def get_new_token(self) -> str:
        """
        Make a POST request to the authentication URL and obtain new auth
        token.

        :raises: :exc:`requests.HTTPError` on HTTP status codes between 400
                 and 599
        :raises: :exc:`~interface.UnexpectedAuthError` on missing token in
                 HTTP response body

        :return: authentication token to be passed with every request
        """
        raise NotImplementedError


class BodyTokenRetrievalStrategy(TokenRetrievalStrategy):
    """
    This strategy knows how to retrieve an authentication token
    contained in the body of the HTTP response from the server.
    """
    def get_new_token(self) -> str:
        response = requests.post(self._url, json=self._credentials)
        response.raise_for_status()
        new_token = response.json().get('token')
        if not new_token:
            raise UnexpectedAuthError(AUTH_ERROR)
        return new_token


class BearerAuth(AuthBase):
    """
    Subclass of :class:`requests.AuthBase`, defining how to authenticate with
    every request via the Bearer authentication protocol defined in RFC 6750.
    """
    def __init__(self, token: str):
        """
        :param token: Authentication token obtained from the auth endpoint
        """
        self._token = token

    def __call__(
        self,
        r: requests.PreparedRequest,
    ) -> requests.PreparedRequest:
        r.headers['Authorization'] = f'Bearer {self._token}'
        return r


class JWTBearerAuth(BearerAuth):
    """
    Subclass of :class:`requests.AuthBase`, defining how to authenticate
    with every request using JSON Web Tokens defined in RFC 7519.
    """
    def __init__(self, token_retrieval_strategy: TokenRetrievalStrategy):
        """
        :param token_retrieval_strategy: means of authentication
        """
        super().__init__('')
        self._token_retrieval_strategy = token_retrieval_strategy

    def __call__(
        self,
        r: requests.PreparedRequest,
    ) -> requests.PreparedRequest:
        now_with_padding = datetime.now() + timedelta(seconds=30)
        if (
            not self._token
            or self.token_expiration(self._token) <= now_with_padding
        ):
            self._token = self._token_retrieval_strategy.get_new_token()
        return super().__call__(r)

    @staticmethod
    def token_expiration(jwt_token) -> datetime:
        """
        Return expiration datetime of given ``jwt_token``.

        :param jwt_token: string representation of a JWT token (encoded)

        :raises: :exc:`~interface.UnexpectedAuthError` if
                 ``jwt_token`` is not a valid JWT token

        :return: datetime of expiration
        """
        try:
            # We won't verify the signature as we do
            # not have the secret from the server.
            jwt_claims = jwt.decode(
                jwt_token,
                options={'verify_signature': False},
            )
        except jwt.DecodeError as err:
            raise UnexpectedAuthError(INVALID_TOKEN_ERROR) from err
        exp = jwt_claims.get('exp', 0)
        return datetime.fromtimestamp(exp)



from dsw_sdk.common.attributes import (
    AttributesMixin,
    DateTimeAttribute,
    ListAttribute,
    ObjectAttribute,
    StringAttribute,
)
from dsw_sdk.common.types import StringType


OUTDATED_PACKAGE_STATE = 'OutdatedPackageState'
UNKNOWN_PACKAGE_STATE = 'UnknownPackageState'
UNPUBLISHED_PACKAGE_STATE = 'UnpublishedPackageState'
UP_TO_DATE_PACKAGE_STATE = 'UpToDatePackageState'
PACKAGE_STATES = (
    OUTDATED_PACKAGE_STATE,
    UNKNOWN_PACKAGE_STATE,
    UNPUBLISHED_PACKAGE_STATE,
    UP_TO_DATE_PACKAGE_STATE,
)


class OrganizationSimple(AttributesMixin):
    logo = StringAttribute(nullable=True)
    name = StringAttribute()
    organization_id = StringAttribute()


class PackageSimpleDTO(AttributesMixin):
    created_at = DateTimeAttribute()
    description = StringAttribute()
    id = StringAttribute()
    km_id = StringAttribute()
    name = StringAttribute()
    organization = ObjectAttribute(OrganizationSimple, nullable=True)
    organization_id = StringAttribute()
    state = StringAttribute(choices=PACKAGE_STATES)
    version = StringAttribute()
    versions = ListAttribute(StringType())


class PackageSimple(AttributesMixin):
    id = StringAttribute()
    name = StringAttribute()
    version = StringAttribute()



"""
General interfaces for HTTP communication -- HTTP client, HTTP response and
exceptions related to the whole process.
"""
from __future__ import annotations

import json
from typing import Any, Dict


class HttpResponse:
    """
    Object representing basic HTTP response that is passed across this SDK.
    Offers only basic methods, but the original response has to be always
    accessible via the :meth:`orig_response` property.
    """
    @property
    def path(self) -> str:
        raise NotImplementedError

    @property
    def orig_response(self) -> Any:
        raise NotImplementedError

    @property
    def text(self) -> str:
        raise NotImplementedError

    def json(self, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError


class HttpError(Exception):
    """
    General HTTP error occurring while communicating with the DSW API.
    Contains HTTP status code, error message and error response from the
    server.
    """
    _default_msg = 'Http error'

    def __init__(self, status_code: int, response: HttpResponse):
        self.status_code = status_code
        self.msg = self._get_message(response)
        self.response = response
        super().__init__(self.msg)

    def __str__(self):
        return f'[{self.status_code}] {self.msg} ({self.response.path})'

    @classmethod
    def _get_message(cls, response: HttpResponse) -> str:
        """
        Based on a content of the ``response``, extracts the error message.
        """
        try:
            resp = response.json()
        except json.JSONDecodeError:
            return response.text

        response_keys = ('message', 'error')
        msg = cls._default_msg

        for key in response_keys:
            if key not in resp:
                continue
            if 'defaultMessage' in resp[key] and 'params' in resp[key]:
                text = resp[key]['defaultMessage']
                params = tuple(resp[key]['params'])
                msg = text % params
            else:
                msg = resp[key]

        return msg


class BadRequestError(HttpError):
    """
    HTTP error with status code 400 Bad request.
    Client sent some invalid request to the server.
    """
    _default_msg = 'Bad request'

    def __init__(self, response: HttpResponse):
        super().__init__(400, response)


class ForbiddenError(HttpError):
    """
    HTTP error with status code 403 Forbidden.
    Client is not authorized to perform such request.
    """
    _default_msg = 'Forbidden'

    def __init__(self, response: HttpResponse):
        super().__init__(403, response)


class NotFoundError(HttpError):
    """
    HTTP error with status code 404 Not found.
    Client requested a resource that was not found on the server.
    """
    _default_msg = 'Not found'

    def __init__(self, response: HttpResponse):
        super().__init__(404, response)


class UnexpectedAuthError(Exception):
    """
    Something unexpected happened when performing the authentication. There
    may be something wrong with the DSW API or it could be just some temporary
    malfunction.
    """


class HttpClient:
    """
    General interface for any HTTP client responsible for communication with
    the DSW API.

    It does not make any assumptions on how the implementation should work or
    which libraries it should leverage. So if you want implement some client
    that is not shipped with this SDK (e.g. for asynchronous communication) or
    you want to use different libraries for HTTP stuff, feel free to subclass
    this interface and pass an instance of your client to the SDK when the
    initialization goes on.
    """

    def __init__(self, *args, **kwargs):
        pass

    def head(self, path: str, **kwargs) -> HttpResponse:
        """
        Sends a HEAD HTTP request.

        :param path: path for the request
        :param kwargs: Optional arguments that the request takes

        :raises: :exc:`HttpError` on HTTP status codes between 400 and 599
                 while authenticating or doing the request itself
        :raises: :exc:`UnexpectedAuthError` on unexpected errors while
                 authenticating

        :return: a response from the server contained
                 in the :class:`HttpResponse` object.
        """
        raise NotImplementedError

    def options(self, path: str, **kwargs) -> HttpResponse:
        """
        Sends a OPTIONS HTTP request.

        :param path: path for the request
        :param kwargs: Optional arguments that the request takes

        :raises: :exc:`HttpError` on HTTP status codes between 400 and 599
                 while authenticating or doing the request itself
        :raises: :exc:`UnexpectedAuthError` on unexpected errors while
                 authenticating

        :return: a response from the server contained
                 in the :class:`HttpResponse` object.
        """
        raise NotImplementedError

    def get(self, path: str, **kwargs) -> HttpResponse:
        """
        Sends a GET HTTP request.

        :param path: path for the request
        :param kwargs: Optional arguments that the request takes

        :raises: :exc:`HttpError` on HTTP status codes between 400 and 599
                 while authenticating or doing the request itself
        :raises: :exc:`UnexpectedAuthError` on unexpected errors while
                 authenticating

        :return: a response from the server contained
                 in the :class:`HttpResponse` object.
        """
        raise NotImplementedError

    def post(self, path: str, body: Dict[str, Any] = None,
             **kwargs) -> HttpResponse:
        """
        Sends a POST HTTP request.

        :param path: path for the request
        :param body: body of the request
        :param kwargs: Optional arguments that the request takes

        :raises: :exc:`HttpError` on HTTP status codes between 400 and 599
                 while authenticating or doing the request itself
        :raises: :exc:`UnexpectedAuthError` on unexpected errors while
                 authenticating

        :return: a response from the server contained
                 in the :class:`HttpResponse` object.
        """
        raise NotImplementedError

    def put(self, path: str, body: Dict[str, Any] = None,
            **kwargs) -> HttpResponse:
        """
        Sends a PUT HTTP request.

        :param path: path for the request
        :param body: body of the request
        :param kwargs: Optional arguments that the request takes

        :raises: :exc:`HttpError` on HTTP status codes between 400 and 599
                 while authenticating or doing the request itself
        :raises: :exc:`UnexpectedAuthError` on unexpected errors while
                 authenticating

        :return: a response from the server contained
                 in the :class:`HttpResponse` object.
        """
        raise NotImplementedError

    def delete(self, path: str, **kwargs) -> HttpResponse:
        """
        Sends a DELETE HTTP request.

        :param path: path for the request
        :param kwargs: Optional arguments that the request takes

        :raises: :exc:`HttpError` on HTTP status codes between 400 and 599
                 while authenticating or doing the request itself
        :raises: :exc:`UnexpectedAuthError` on unexpected errors while
                 authenticating

        :return: a response from the server contained
                 in the :class:`HttpResponse` object.
        """
        raise NotImplementedError
